import numpy as np
import logging
from fitxf.math.fit.cluster.Cluster import Cluster
from fitxf.math.utils.Logging import Logging


class ClusterUnitTest:

    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()

    def test_converge(self):
        x = np.array([
            [5, 1, 1], [8, 2, 1], [6, 0, 2],
            [1, 5, 1], [2, 7, 1], [0, 6, 2],
            [1, 1, 5], [2, 1, 8], [0, 2, 6],
        ])
        obj = Cluster(logger=self.logger)
        res = obj.kmeans_optimal(
            x = x,
            estimate_min_max = True,
        )
        n_iters = res[0]['total_iterations']
        n = res[0]['n_centers']
        centers = res[0]['cluster_centers']
        center_lbls = res[0]['cluster_labels']
        self.logger.info('Total iterations from full train = ' + str(n_iters))

        assert n == 3, 'Expect 3 centers but got ' + str(n)

        for i in range(len(x)):
            obs = center_lbls[i]
            exp = center_lbls[i-i%3]
            assert obs == exp, \
                'Label for index ' + str(i) + ', x = ' + str(x[i]) + ' observed ' + str(obs) + ', expected ' + str(exp)

        #
        # Test fine-tuning
        #
        res = obj.kmeans(
            x = x,
            n_centers = n,
            start_centers = centers,
        )
        n_iters = res['total_iterations']
        new_centers = res['cluster_centers']
        new_center_lbls = res['cluster_labels']
        self.logger.info('Total iterations from fine tuning SAME DATA = ' + str(n_iters))
        assert n_iters == 1, 'Fine tuning same data should have only 1 iteration but got ' + str(n_iters)
        diff_centers = np.sum( (new_centers - centers)**2 )
        assert diff_centers < 0.0000000001, \
            'Fine tuning same data should not change centers:\n' + str(centers) + '\nbut changed to:\n' + str(new_centers)

        # add new point
        x_add = np.array([[7, 1, 1]])
        x_new = np.append(x, x_add, axis=0)
        res = obj.kmeans(
            x = x_new,
            n_centers = n,
            start_centers = centers,
        )
        n_iters = res['total_iterations']
        new_enters = res['cluster_centers']
        new_center_lbls = res['cluster_labels']
        self.logger.info('Total iterations from fine tuning with new data = ' + str(n_iters))

        # Last added point cluster number must equal 1st one
        assert new_center_lbls[-1] == new_center_lbls[0], \
            'Last added point should belong to cluster of 1st 3 points but got labels ' + str(new_center_lbls)
        # This is same as above, test that original clusters retained
        for i in range(len(x_new) - 1):
            obs = new_center_lbls[i]
            exp = new_center_lbls[i-i%3]
            assert obs == exp, \
                'Label for index ' + str(i) + ', x = ' + str(x[i]) + ' observed ' + str(obs) + ', expected ' + str(exp)

        return

    def test_diverge(self):
        x = np.array([
            [0, 0, 0, 0, 1], [0, 0, 0, 1, 0], [0, 0, 1, 0, 0], [0, 1, 0, 0, 0], [1, 0, 0, 0, 0],
        ])
        obj = Cluster(logger=self.logger)
        res = obj.kmeans_optimal(
            x = x,
            # allow single point clusters
            thr_single_clusters = 1.,
            estimate_min_max    = False,
            max_clusters        = len(x),
        )
        n = res[0]['n_centers']
        centers = res[0]['cluster_centers']
        center_lbls = res[0]['cluster_labels']
        obj.logger.debug('Optimal clusters = ' + str(n))
        obj.logger.debug('Cluster centers = ' + str(centers))
        obj.logger.debug('Cluster labels: ' + str(center_lbls))
        obj.logger.debug('Cluster sizes: ' + str(res[0]['cluster_sizes']))
        assert n >= 4

    def test_imbalanced(self):
        x = np.array([
            [1.0, 0, 0, 0, 0], [1.1, 0, 0, 0, 0], [0, 0, 0, 0, 1.0], [0, 0, 0, 0, 1.1], [0, 0, 0, 0, 0.9],
            [100, 0, 0, 0, 0], [101, 0, 0, 0, 0], [0, 0, 0, 0, 100], [0, 0, 0, 0, 110], [0, 0, 0, 0, 99],
        ])
        obj = Cluster(logger=self.logger)
        res = obj.kmeans_optimal(
            x = x,
            estimate_min_max = True,
        )
        # TODO Why index 1 and not 0?
        n = res[1]['n_centers']
        centers = res[1]['cluster_centers']
        center_lbls = res[1]['cluster_labels']
        obj.logger.debug('Optimal clusters = ' + str(n))
        obj.logger.debug('Cluster centers = ' + str(centers))
        obj.logger.debug('Cluster labels: ' + str(center_lbls))
        obj.logger.debug('Cluster sizes: ' + str(res[1]['cluster_sizes']))
        assert n == 4

    def test_pass_thru_mode(self):
        x = np.array([
            [5, 1, 1], [8, 2, 1], [6, 0, 2],
            [1, 5, 1], [2, 7, 1], [0, 6, 2],
            [1, 1, 5], [2, 1, 8], [0, 2, 6],
        ])
        x_labels = ['a', 'a', 'a', 'b', 'b', 'b', 'c', 'c', 'c']
        expected_cluster_labels = list(range(len(x)))
        obj = Cluster(logger=self.logger)
        res = obj.kmeans_optimal(
            x = x,
            x_labels = x_labels,
            estimate_min_max = True,
            test_mode = True,
        )
        n_iters = res[0]['total_iterations']
        n = res[0]['n_centers']
        centers = res[0]['cluster_centers']
        cluster_lbls = res[0]['cluster_labels']
        self.logger.info('Total iterations from full train = ' + str(n_iters))
        assert n == len(x), \
            'n centers  should be just length of x ' + str(len(x)) + ', but got ' + str(n)
        assert np.sum((centers - x)**2) < 0.0000000001, \
            'Centers should just be x but centers ' + str(centers) + ', x ' + str(x)
        assert cluster_lbls == expected_cluster_labels, \
            'Center labels should be ' + str(expected_cluster_labels) + ', got ' + str(cluster_lbls)
        return

    def test(self):
        self.test_converge()
        self.test_diverge()
        self.test_imbalanced()
        self.test_pass_thru_mode()
        print('ALL TESTS PASSED OK')
        return


if __name__ == '__main__':
    ut = ClusterUnitTest(
        logger = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    )
    ut.test()
    exit(0)



import logging
import os
import platform
import numpy as np
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Profile import Profiling


# Important Note:
#     Logging of huge pandas/numpy arrays commented out in release version
#     as it slows down code by hundreds of milliseconds.
class MathUtils:

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        self.enable_slow_logging_of_numpy = os.environ.get("ENABLE_SLOW_LOGGING", "false").lower() in ["true", "1"]
        return

    def match_template_1d(
            self,
            x: np.ndarray,
            seq: np.ndarray,
    ) -> dict:
        x = np.array(x) if type(x) in (list, tuple) else x
        seq = np.array(seq) if type(seq) in (list, tuple) else seq
        assert x.ndim == seq.ndim == 1
        assert len(seq) <= len(x)

        l_x, l_seq = len(x), len(seq)

        # Template for sequence
        r_seq = np.arange(l_seq)

        # Create the template matching indices in 2D. e.g.
        #   [
        #     [0],
        #     [1],
        #     [2],
        #     ...
        #     [n],
        #     ...
        #   ]
        template_matching_indices = np.arange(l_x - l_seq + 1)[:, None]
        if self.enable_slow_logging_of_numpy:
            self.logger.debug('Template matching indices 2D structure: ' + str(template_matching_indices))
        # Create the template matching indices in 2D. e.g. for seq length 3
        #   [
        #     [0,1,2],
        #     [1,2,3],
        #     [2,3,4],
        #     ...
        #     [n,n+1,n+2],
        #     ...
        #   ]
        template_matching_indices = template_matching_indices + r_seq
        if self.enable_slow_logging_of_numpy:
            self.logger.debug('Template matching indices final: ' + str(template_matching_indices))
        # Find matches
        template_matches = x[template_matching_indices] == seq
        if self.enable_slow_logging_of_numpy:
            self.logger.debug('Template matches for seq ' + str(seq) + ': ' + str(template_matches))

        #
        # nan means "*" match like in string regex
        #
        nan_positions = np.isnan(seq)
        if self.enable_slow_logging_of_numpy:
            self.logger.debug('nan positions: ' + str(nan_positions))
        template_matches = 1 * (template_matches | nan_positions)
        if self.enable_slow_logging_of_numpy:
            self.logger.debug('Template matches with nan for seq ' + str(seq) + ': ' + str(template_matches))

        # Match is when all are 1's
        match_start_indexes = 1 * (np.sum(template_matches, axis=-1) == len(seq))
        if self.enable_slow_logging_of_numpy:
            self.logger.debug('Match start indexes: ' + str(match_start_indexes))
        match_all_indexes = np.convolve(match_start_indexes, np.ones(l_seq, dtype=int), mode='full')
        if self.enable_slow_logging_of_numpy:
            self.logger.debug('Match all indexes  : ' + str(match_all_indexes))

        # Get the range of those indices as final output
        if match_start_indexes.any() > 0:
            # res =  np.argwhere(match_start_indexes == 1).flatten().tolist()
            res = {
                'match_indexes': np.argwhere(
                    match_start_indexes == 1
                ).flatten().tolist(),
                'match_sequence': np.argwhere(
                    np.convolve(match_start_indexes, np.ones((l_seq), dtype=int) > 0)
                ).flatten().tolist()
            }
        else:
            # res = []
            res = {
                'match_indexes': [],
                'match_sequence': [],  # No match found
            }
        return res

    def match_template(
            self,
            x: np.ndarray,
            seq: np.ndarray,
            seq_real_shape: list = None,
            return_only_start_indexes = True,
    ) -> dict:
        x = np.array(x) if type(x) in (list, tuple) else x
        seq = np.array(seq) if type(seq) in (list, tuple) else seq
        assert x.ndim == seq.ndim, 'Dimensions do not match, x dim ' + str(x.ndim) + ', seq dim ' + str(seq.ndim)
        n_dim = x.ndim

        # Convert to ndim, same as converting to a base-N number
        if n_dim > 1:
            assert seq_real_shape is not None
            x_1d = x.flatten()
            seq_1d = seq.flatten()
            # count_valid = np.sum(1 * np.logical_not(np.isnan(seq_1d)))
            # Remove ending nan(s)
            for i in range(len(seq_1d)):
                if np.isnan(seq_1d[-1]):
                    seq_1d = seq_1d[:-1]
            # if self.enable_slow_logging_of_numpy:
            #     self.logger.debug('Sequence flattened ' + str(seq_1d))

            res = self.match_template_1d(x=x_1d, seq=seq_1d)
            match_start_indexes_1d, match_seq_1d = res['match_indexes'], res['match_sequence']
            if self.enable_slow_logging_of_numpy:
                self.logger.debug('Match 1d result ' + str(res))

            bases = list(x.shape) + [1]
            match_indexes = []
            match_sequence = []
            len_cycle = int(len(match_seq_1d) / len(match_start_indexes_1d))
            if self.enable_slow_logging_of_numpy:
                self.logger.debug(
                    'Length of cycle ' + str(len_cycle) + ' for match start indexes 1d ' + str(match_start_indexes_1d)
                    + ', match seq 1d: ' + str(match_seq_1d)
                )

            for i, idx_1d in enumerate(match_start_indexes_1d):
                nbr_rep = self.convert_to_multibase_number(n=idx_1d, bases=bases, min_digits=x.ndim)
                border_points = np.array(nbr_rep) + np.array(seq_real_shape) - 1
                check_if_overflow_x = np.min(np.array(x.shape) - 1 - border_points) < 0
                if check_if_overflow_x:
                    if self.enable_slow_logging_of_numpy:
                        self.logger.debug(
                            'Start index IGNORE ' + str(nbr_rep) + ', seq: ' + str(seq)
                            + ', shape real ' + str(seq_real_shape) + ', border points ' + str(border_points)
                            + ', check overflow ' + str(check_if_overflow_x)
                        )
                    continue
                # else:
                #     if self.enable_slow_logging_of_numpy:
                #         self.logger.debug(
                #             'Start index OK ' + str(nbr_rep) + ', seq: ' + str(seq)
                #             + ', shape real ' + str(seq_real_shape) + ', border points ' + str(border_points)
                #             + ', check overflow ' + str(check_if_overflow_x)
                #         )

                if self.enable_slow_logging_of_numpy:
                    self.logger.debug(
                        'Match indexes converted idx ' + str(idx_1d) + ' to base ' + str(bases) + ' number: ' + str(nbr_rep)
                    )
                i_start = i*len_cycle
                match_seq_1d_1cycle = match_seq_1d[i_start:(i_start + len_cycle)]
                if self.enable_slow_logging_of_numpy:
                    self.logger.debug(
                        'Match seq 1d 1 cycle at index ' + str(i_start) + ': ' + str(match_seq_1d_1cycle)
                    )
                coor = self.get_coordinates(
                    x = x,
                    match_seq_1d_1cycle = match_seq_1d_1cycle,
                    seq_original = seq,
                )
                # if is_valid:
                match_indexes.append(nbr_rep)
                match_sequence = match_sequence + coor
                # else:
                #     if self.enable_slow_logging_of_numpy:
                #         self.logger.info(
                #             'Invalid match sequence ' + str(match_seq_1d_1cycle) + ', coordinates ' + str(coor)
                #         )
            return match_indexes if return_only_start_indexes else \
                {'match_indexes': match_indexes, 'match_sequence': match_sequence}
        else:
            res = self.match_template_1d(x=x, seq=seq)
            return res['match_indexes'] if return_only_start_indexes else res

    def convert_to_multibase_number(
            self,
            n: int,      # base 10 number
            bases: list,   # base to convert to, e.g. [6, 11, 1] --> last digit always 1
            min_digits: int = 0,
    ):
        assert n >= 0
        nbr_rep = []
        base = 1
        for idx in range(len(bases)-1):
            base *= bases[-(idx+2)]
            remainder = int(n % base)
            nbr_rep.append(remainder)
            n = (n - remainder) / base
            # if self.enable_slow_logging_of_numpy:
            #     self.logger.debug('idx=' + str(idx) + ', base=' + str(base) + ', remainder=' + str(remainder))
        if n > 0:
            nbr_rep.append(n)
        while len(nbr_rep) < min_digits:
            nbr_rep.append(0)
        nbr_rep.reverse()
        return nbr_rep

    def get_coordinates(
            self,
            x: np.ndarray,
            match_seq_1d_1cycle: list,
            seq_original: np.ndarray,
    ) -> (list, bool):
        assert len(match_seq_1d_1cycle) > 0
        bases = list(x.shape) + [1]
        coordinates = []
        for idx in match_seq_1d_1cycle:
            nbr_rep = self.convert_to_multibase_number(n=idx, bases=bases, min_digits=x.ndim)
            coordinates.append(nbr_rep)
            if self.enable_slow_logging_of_numpy:
                self.logger.debug(
                    'Match sequence converted idx ' + str(idx) + ' to base ' + str(bases) + ' number: ' + str(nbr_rep)
                )
        if self.enable_slow_logging_of_numpy:
            self.logger.debug('Converted match seq ' + str(match_seq_1d_1cycle) + ' to ' + str(coordinates))
        # shape_len_1d = len(match_seq_1d_1cycle)
        # self.logger.info('Shape len 1d ' + str(shape_len_1d) + ' for ' + str(seq))
        # Check if result spans across row/column/etc borders, which is not valid
        # span_min = np.min(np.array(coordinates[:shape_len_1d]), axis=0)
        # span_max = np.max(np.array(coordinates[:shape_len_1d]), axis=0)
        # span_shape = span_max - span_min + 1
        # seq_shape = np.array(list(seq_original.shape))
        # self.logger.info('Span min: ' + str(span_min) + ' for ' + str(match_sequence[:shape_len_1d]))
        # self.logger.info('Span max: ' + str(span_max) + ' for ' + str(match_sequence[:shape_len_1d]))
        # self.logger.info(
        #     'Span shape ' + str(span_shape) + ', seq shape ' + str(seq_shape) + ', diff ' + str(span_shape - seq_shape)
        # )
        # not_cross_rows = np.max(span_shape - seq_shape) <= 0
        # valid = not_cross_rows
        return coordinates

    def sample_random_no_repeat(
            self,
            list,
            n,
    ):
        assert n <= len(list)
        rng = np.random.default_rng()
        numbers = rng.choice(len(list), size=n, replace=False)
        sampled = []
        for i in numbers:
            sampled.append(list[i])
        return sampled


class MathUtilsUnitTest:
    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        self.mu = MathUtils(logger=self.logger)
        return

    def test(self):
        for n, bases, exp in [
            (19, [5, 1], [3, 4]),
            (29, [13, 1], [2, 3]),
            (38, [13, 1], [2, 12]),
            (0, [3, 2, 1], [0, 0]),
            (1, [3, 2, 1], [0, 1]),
            (5, [3, 2, 1], [2, 1]),
            (5, [3, 2, 1, 1], [2, 1, 0]),
            (5, [3, 2, 1, 1, 1], [2, 1, 0, 0]),
            (29, [30, 20, 1], [1, 9]),
        ]:
            res = self.mu.convert_to_multibase_number(n=n, bases=bases, min_digits=len(bases)-1)
            assert res == exp, \
                'Test base convertion n=' + str(n) + ', bases=' + str(bases) + ', exp=' + str(exp) + ', res=' + str(res)

        self.test_1d()
        self.logger.info('1-DIMENSION TESTS PASSED')
        self.test_2d()
        self.logger.info('2-DIMENSION TESTS PASSED')

        os.environ["ENABLE_SLOW_LOGGING"] = "false"
        rps_no_logging = self.test_speed()

        os.environ["ENABLE_SLOW_LOGGING"] = "true"
        try:
            self.test_speed()
        except Exception as ex:
            self.logger.info(
                'Expected to fail when slow logging enabled (RPS no logging=' + str(rps_no_logging) + '): ' + str(ex)
            )
        self.logger.info('SPEED TESTS PASSED')

        self.logger.info('ALL TESTS PASSED')
        return

    def test_1d(self):
        # Test 1D
        # array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        x = np.arange(20) % 10
        for i, (seq, exp_matches, exp_seq) in enumerate([
            (np.array([1, 2, 3, 4]), np.array([1, 11]), np.array([1,  2,  3,  4, 11, 12, 13, 14])),
            (np.array([1, np.nan, np.nan, 4]), np.array([1, 11]), np.array([1,  2,  3,  4, 11, 12, 13, 14])),
            (np.array([9, 10, 11]), np.array([]), np.array([])),
            # (np.array([1, 3, 5]), []),
        ]):
            res = self.mu.match_template(
                x = x,
                seq = seq,
                return_only_start_indexes = False,
            )
            match_idxs, match_seq = res['match_indexes'], res['match_sequence']
            self.logger.info('Test result 1D #' + str(i) + ': ' + str(res))
            assert np.sum((np.array(match_idxs) - exp_matches)**2) < 0.0000000001, \
                '#' + str(i) + ' Match indexes ' + str(match_idxs) + ' not ' + str(exp_matches)
            assert np.sum((np.array(match_seq) - exp_seq)**2) < 0.0000000001, \
                '#' + str(i) +' Match sequence ' + str(match_seq) + ' not ' + str(exp_seq)
        return

    def test_2d(self):
        # Test 2D
        # [[0 1 2 3 4]
        #  [5 6 7 8 9]
        #  [0 1 2 3 4]
        #  [5 6 7 8 9]]
        x = np.arange(20) % 10
        x.resize((4, 5))
        nan = np.nan
        self.logger.info('2D test data:\n' + str(x))

        for i, (seq, seq_ori_shape, exp_matches, exp_seq) in enumerate([
            (np.array([[1, 2, nan, nan, nan], [6, 7, nan, nan, nan]]), (2, 2), np.array([[0,1], [2,1]]),
             np.array([[0, 1], [0, 2], [0, 3], [0, 4], [1, 0], [1, 1], [1, 2], [2, 1], [2, 2], [2, 3], [2, 4], [3, 0], [3, 1], [3, 2]])),
            # (np.array([[1, 2], [6, 7]]), np.array([[0, 1], [2, 1]]), np.array([])),
            # Test invalid of cross rows
            (np.array([[4, 5, nan, nan, nan], [9, 0, nan, nan, nan]]), (2, 2), np.array([]),
             np.array([])),
        ]):
            res = self.mu.match_template(
                x = x,
                seq = seq,
                seq_real_shape = seq_ori_shape,
                return_only_start_indexes = False,
            )
            match_idxs, match_seq = res['match_indexes'], res['match_sequence']
            self.logger.info('Test result 2D #' + str(i) + ': ' + str(res))
            assert np.sum((np.array(match_idxs) - exp_matches)**2) < 0.0000000001, \
                '#' + str(i) + ' Match indexes ' + str(match_idxs) + ' not ' + str(exp_matches)
            assert np.sum((np.array(match_seq) - exp_seq) ** 2) < 0.0000000001, \
                '#' + str(i) + ' Match sequence ' + str(match_seq) + ' not ' + str(exp_seq)
        return

    def test_speed(self):
        new_obj = MathUtils(logger=self.logger)
        profiler = Profiling(logger=self.logger)
        x = np.arange(20) % 10
        x.resize((4, 5))
        nan = np.nan
        start_time = profiler.start()
        n = 1000
        for i in range(n):
            _ = new_obj.match_template(
                x = x,
                seq = np.array([[1, 2, nan, nan, nan], [6, 7, nan, nan, nan]]),
                seq_real_shape = [2, 2],
                return_only_start_indexes = False,
            )
        diffsecs = profiler.get_time_dif_secs(start=start_time, stop=profiler.stop())
        rps = round(n / diffsecs, 3)
        msec_avg = round(1000 * diffsecs / n, 3)
        self.logger.info(
            'RPS match template n=' + str(n) + ', total secs=' + str(diffsecs) + ', rps=' + str(rps)
            + ', msec avg=' + str(msec_avg)
        )
        pf = platform.platform()
        pf_info = pf.split("-")
        processor = pf_info[2]
        if processor in ['x86_64']:
            rps_thr, msec_thr = 1000, 0.5
        else:
            rps_thr, msec_thr = 10000, 0.1
        self.logger.info(
            'Platform processor "' + str(processor) + '", "' + str(pf) + '": ' + str(pf_info)
            + ', rps thr ' + str(rps_thr)
        )

        assert rps > rps_thr, \
            'FAILED RPS n=' + str(n) + ', total=' + str(diffsecs) + 's, rps=' + str(rps)
        assert msec_avg < msec_thr, \
            'FAILED RPS n=' + str(n) + ', total=' + str(diffsecs) + 's, msec avg=' + str(msec_avg)
        return rps


if __name__ == '__main__':
    # os.environ["ENABLE_SLOW_LOGGING"] = "true"
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    MathUtilsUnitTest(logger=lgr).test()
    mu = MathUtils(logger=lgr)
    res = mu.sample_random_no_repeat(
        list = np.arange(100).tolist() + np.arange(100).tolist(),
        n = 100,
    )
    res.sort()
    print(res)
    exit(0)



import logging
import traceback
import numpy as np
import pandas as pd
from fitxf import MathUtils
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Profile import Profiling
from fitxf.math.utils.Pandas import Pandas


class PatternSearch:

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()

        self.profiler = Profiling(logger=self.logger)
        self.math_utils= MathUtils(logger=self.logger)
        return

    def heuristic_guess_start_search_locations(
            self,
            x: np.ndarray,
            possible_start_values: np.ndarray,
            seq_len: int,
            top_k: int = 10,
    ):
        assert seq_len <= len(x), 'Sequence length must be <= ' + str(len(x))
        start_time = self.profiler.start()

        # Find indexes with the possible start values
        test_diff = x[:,None] - possible_start_values
        # self.logger.debug('Test diff ' + str(test_diff))
        indexes_possible = np.argwhere(np.prod(test_diff, axis=-1) == 0).flatten()
        # Move one character to the next character, since these are separators
        indexes_possible = np.minimum(len(x)-1, indexes_possible + 1)
        # self.logger.debug(
        #     'Possible indexes ' + str(indexes_possible) + ' at ' + str([chr(v) for v in x[indexes_possible]])
        # )

        successful_indexes = []
        hint_prefix_indexes = {}
        for i in indexes_possible:
            if i + seq_len > len(x):
                continue
            if i in successful_indexes:
                continue
            seq_start, seq_end = i, i + seq_len
            seq_i = x[seq_start:seq_end]
            seq_i_str = ''.join([chr(v) for v in seq_i])

            match_indexes = self.math_utils.match_template(
                x = x,
                seq = seq_i,
                return_only_start_indexes = True,
            )
            success = len(match_indexes) >= 2
            if not success:
                continue
            else:
                hint_prefix_indexes[seq_i_str] = hint_prefix_indexes.get(seq_i_str, []) + match_indexes
                successful_indexes += match_indexes
                self.logger.debug(
                    'Match indexes at index ' + str(seq_start) + ' for seq len ' + str(seq_len) + ' "'
                    + str(''.join([chr(v) for v in seq_i])) + '": ' + str(match_indexes) + ', hint indexes now: '
                    + str(hint_prefix_indexes)
                )

        successful_indexes.sort()
        self.logger.info('Successful hint indexes: ' + str(successful_indexes))

        if len(hint_prefix_indexes) > 0:
            # sort and select top_k
            records = [
                {'prefix': prefix, 'indexes': indexes, 'count': len(indexes)}
                for prefix, indexes in hint_prefix_indexes.items()
            ]
            df_records = pd.DataFrame.from_records(records)
            df_records = df_records.sort_values(by=['count'], ascending=False)
            df_records = df_records[0:min(top_k, len(df_records))]
            self.logger.info('Final sorted hint prefix indexes: ' + str(df_records))
            hint_prefix_indexes_sorted = {r['prefix']: r['indexes'] for r in df_records.to_dict(orient='records')}
        else:
            hint_prefix_indexes_sorted = {}

        diff_msecs = round(1000 * self.profiler.get_time_dif_secs(start=start_time, stop=self.profiler.stop()), 2)
        self.logger.info('Hint indexes search for ' + str(possible_start_values) + ' took ' + str(diff_msecs) + 'ms')

        return hint_prefix_indexes_sorted

    def repeat_density_and_coverage(
            self,
            match_indexes: list,
            seq_len: int,
            total_len: int,
    ):
        i_start = np.min(match_indexes)
        i_end = min(np.max(match_indexes) + seq_len, total_len)
        # Density of repeat sequences in the range of valid repeats
        density = round(len(match_indexes) * seq_len / (i_end - i_start), 3)
        # Coverage of the repeat range in the whole length
        coverage = round((i_end - i_start) / total_len, 3)
        return density, coverage

    def find_repeat_sequences(
            self,
            x: np.ndarray,
            hint_separators: np.ndarray,
            density_repeat_thr: float = 0.8,
            coverage_thr: float = 0.5,
            # roughly 10 words
            min_seq_len: int = 64,
            top_k_hints: int = 10,
            stagnant_at_optimal_count_thr: int = 5,
            string_convert_result: bool = False,
    ):
        start_time = self.profiler.start()

        hint_prefix_indexes = self.heuristic_guess_start_search_locations(
            x = x,
            possible_start_values = hint_separators,
            seq_len = min_seq_len,
            top_k = top_k_hints,
        )

        repeat_indexes_and_texts = []
        for prefix, indexes in hint_prefix_indexes.items():
            match_indexes_start = np.array(indexes)
            intervals = match_indexes_start[1:] - match_indexes_start[:-1]
            max_interval = np.min(intervals)
            min_interval = min_seq_len
            if min_interval > max_interval:
                continue
            seq_list = [min_interval]
            self.logger.info(
                'Using min interval ' + str(min_seq_len) + ', max interval ' + str(max_interval)
                + ' for index intervals: ' + str(intervals)
            )
            ind_optimal = None
            break_reason = None
            stagnant_count = 0
            while True:
                # TODO Don't just take the first one
                # Take the first one, the last one might be too short and will mistakenly cut the match sequence
                idx_start = match_indexes_start[0]
                idx_end = min(idx_start + seq_list[-1], len(x))
                seq_i = x[idx_start:idx_end]
                self.logger.debug(
                    'Match template for idx [' + str(idx_start) + ',' + str(idx_end) + '], seq len ' + str(seq_list[-1])
                    + ' "' + str(''.join([chr(v) for v in seq_i])) + '"'
                )
                match_indexes_end_tmp = self.math_utils.match_template(
                    x = x,
                    seq = seq_i,
                    return_only_start_indexes = True,
                )
                self.logger.debug(
                    'Match indexes for seq len ' + str(seq_list[-1]) + ' "' + str(''.join([chr(v) for v in seq_i]))
                    + '": ' + str(match_indexes_end_tmp)
                )

                is_success_match_seq = len(match_indexes_end_tmp) >= 2
                improve_from_prev_optimal = False

                #
                # Check to see if we got new optimal if success in matching
                #
                if is_success_match_seq:
                    ind_tmp = self.get_repeat_indicators(
                        match_indexes = match_indexes_end_tmp,
                        seq_list = seq_list,
                        x = x,
                        string_convert = string_convert_result,
                    )
                    density, coverage, score = ind_tmp["density"], ind_tmp["coverage"], ind_tmp["score"]
                    density_cond_met = density >= density_repeat_thr
                    coverage_cond_met = coverage >= coverage_thr
                    prev_score = ind_optimal['score'] if ind_optimal is not None else -np.inf
                    improve_from_prev_optimal = (density_cond_met and coverage_cond_met) and (score > prev_score)

                    if improve_from_prev_optimal:
                        self.logger.info(
                            'Found new optimal from prev ' + str(ind_optimal) + '. Improved: ' + str(ind_tmp)
                        )
                        ind_optimal = ind_tmp
                    else:
                        tmp_prev_optimal = {k: v for k, v in ind_optimal.items() if k not in ["s_parts", "prefix"]} \
                            if ind_optimal is not None else None
                        self.logger.info(
                            'No improve from previous optimal ' + str(tmp_prev_optimal) + ' for: '
                            + str({k: v for k, v in ind_tmp.items() if k not in ["s_parts", "prefix"]})
                        )

                    # if (seq_list[-1] == max_interval):
                    #     self.logger.debug(
                    #         'Found optimal by early condition, max seq len ' + str(seq_list)
                    #         + ', max interval ' + str(max_interval)
                    #     )
                    #     break_reason = 'hit max interval'
                    #     break
                    if improve_from_prev_optimal:
                        self.logger.info(
                            'Found optimal by early condition, max seq len ' + str(seq_list)
                            + ', max interval ' + str(max_interval) + ', density ' + str(density)
                            + ' (thr=' + str(density_repeat_thr) + ', cond=' + str(density_cond_met)
                            + '), coverage ' + str(coverage) + '(thr=' + str(coverage_thr)
                            + ', cond=' + str(coverage_cond_met) + ')'
                        )
                        break_reason = 'density/coverage condition met'
                        break
                    elif stagnant_count >= stagnant_at_optimal_count_thr:
                        self.logger.info(
                            'Stagnant for last ' + str(stagnant_count) + ' tries >= '
                            + str(stagnant_at_optimal_count_thr)
                        )
                        break_reason = 'stagnant ' + str(stagnant_count)
                        break

                #
                # Adjust new max/min intervals first
                #
                if not is_success_match_seq:
                    # The last seq len had no match, means we know now the upper bound is at most 1 less
                    tmp = max_interval
                    max_interval = min(max_interval, seq_list[-1] - 1)
                    self.logger.debug(
                        'No match at seq len ' + str(seq_list[-1]) + ', max interval changed from ' + str(tmp)
                        + ' --> ' + str(max_interval)
                    )
                else:
                    # The last seq len search successful, means we know lower bound is at least seq len + 1
                    tmp = min_interval
                    # if improve_from_prev_optimal:
                    #     # Be more daring, move lower bound up by at least half the interval
                    #     min_interval = max(min_interval, int((min_interval + max_interval) / 2))
                    # else:
                    # No change or just add 1 to lower bound, bcos was not optimal
                    min_interval = min_interval+1 if seq_list[-1]==min_interval else min_interval
                    self.logger.debug(
                        'Have match at seq len ' + str(seq_list[-1])
                        + ', min interval changed (or possibly remained) from ' + str(tmp)
                        + ' --> ' + str(min_interval) + ', upper bound ' + str(max_interval)
                        + '. Improve from prev optimal ' + str(improve_from_prev_optimal)
                    )

                remaining_indexes_to_try = np.arange(max_interval - min_interval + 1) + min_interval
                remaining_indexes_to_try = list(
                    set(remaining_indexes_to_try.tolist()).difference(set(seq_list))
                )
                self.logger.debug('Remaining indexes to try: ' + str(remaining_indexes_to_try))
                if not remaining_indexes_to_try:
                    self.logger.debug(
                        'Cannot decrease/increase seq len anymore, min interval=' + str(min_interval)
                        + ', max=' + str(max_interval) + '. No more remaining indexes available in history '
                        + str(seq_list)
                    )
                    break_reason = 'no more valid indexes to try'
                    break

                #
                # Adjust next move
                #
                if not is_success_match_seq:
                    stagnant_count = 0
                    # no match, decrease seq len by approx. half
                    seq_len_new = remaining_indexes_to_try[int(len(remaining_indexes_to_try) / 2)]
                    self.logger.info(
                        'No success, moved seq len from ' + str(seq_list[-1]) + ' --> ' + str(seq_list)
                        + ', lower/upper bound [' + str(min_interval) + ', ' + str(max_interval) + ']'
                    )
                    seq_list.append(seq_len_new)
                elif not improve_from_prev_optimal:
                    stagnant_count += 1
                    seq_len_new = np.random.choice(remaining_indexes_to_try)
                    self.logger.info(
                        'Did not improve optimal, moved seq len randomly from ' + str(seq_list[-1])
                        + ' --> ' + str(seq_len_new)
                        + ', lower/upper bound [' + str(min_interval) + ', ' + str(max_interval) + ']'
                    )
                    seq_list.append(seq_len_new)
                else:
                    stagnant_count = 0
                    # positive match, increase seq len
                    seq_len_new = np.max(remaining_indexes_to_try)
                    self.logger.info(
                        'Success with optimal, moved seq len from ' + str(seq_list[-1]) + ' --> ' + str(seq_len_new)
                    )
                    seq_list.append(seq_len_new)

            if ind_optimal is not None:
                ind_optimal['break_reason'] = break_reason
                repeat_indexes_and_texts.append(ind_optimal)
                self.logger.debug('Appended: ' + str(ind_optimal))
        try:
            df_repeat_indexes = pd.DataFrame.from_records(repeat_indexes_and_texts)
            df_repeat_indexes = df_repeat_indexes.sort_values(by=['score'], ascending=False)
            self.logger.debug('Repeat indexes: ' + str(df_repeat_indexes))
            final_matches = df_repeat_indexes.to_dict(orient='records')
        except Exception as ex:
            self.logger.debug('Possibly no matches found: ' + str(ex) + ' Stack trace: ' + str(traceback.format_exc()))
            final_matches = None

        diff_msecs = round(1000 * self.profiler.get_time_dif_secs(start=start_time, stop=self.profiler.stop()), 2)
        self.logger.info(
            'Repeat sequence search took ' + str(diff_msecs) + 'ms. Sequence lengths in iteration: '
        )
        return final_matches

    def get_repeat_indicators(
            self,
            match_indexes: list,
            seq_list: list,
            x: np.ndarray,
            string_convert: bool = False,
    ):
        if string_convert:
            string = ''.join([chr(v) for v in x])
        else:
            string = x
        # self.logger.debug('Match indexes end: ' + str(match_indexes))
        # self.logger.debug('Match indexes end: ' + str([string[i:(i + 50)] for i in match_indexes]))
        seq_len = seq_list[-1]
        i_start = np.min(match_indexes)
        i_end = min(np.max(match_indexes) + seq_len, len(x))
        s_parts = [string[:i_start], string[i_start:i_end], string[i_end:]]
        prefix = string[i_start:(i_start+seq_len)]
        # self.logger.debug('New prefix "' + str(prefix) + '"')

        density, coverage = self.repeat_density_and_coverage(
            match_indexes = match_indexes,
            seq_len = seq_len,
            total_len = len(string)
        )
        # self.logger.debug(
        #     'Seq len ' + str(seq_len) + ' "' + str(string[i_start:(i_start + seq_len)])
        #     + '", start ' + str(i_start) + ', end ' + str(i_end)
        #     + ', density of repeats ' + str(density) + ', coverage ' + str(coverage)
        # )

        tmp = np.array(match_indexes)
        intervals = tmp[1:] - tmp[:-1]
        return {
            'indexes': match_indexes,
            'seq_len': seq_len,
            'seq_list': seq_list,
            'prefix': prefix,
            'index_start': i_start,
            'index_end': i_end,
            'min_interval': np.min(intervals),
            'max_interval': np.max(intervals),
            # 'texts': [string[i:(i+seq_len_end)] for i in match_indexes_end],
            'density': density,
            'coverage': coverage,
            'score': density * coverage,
            's_parts': s_parts,
        }


if __name__ == '__main__':
    Pandas.increase_display()
    lgr = Logging.get_default_logger(log_level=logging.DEBUG, propagate=False)
    ps = PatternSearch(logger=lgr)
    res = ps.find_repeat_sequences(
        x = np.array([
            ord(c) for c in "pattern 1 pattern 2 pattern 3 pattern 4 ..."]
        ),
        min_seq_len = 4,
        hint_separators = np.array([ord(c) for c in ["1", " "]]),
        string_convert_result = True,
        density_repeat_thr = 0.5,
        coverage_thr = 0.5,
    )
    print(res)
    exit(0)



import logging
import threading
import time
import re
import numpy as np
from datetime import datetime
from fitxf.math.utils.Logging import Logging


#
# Proper wrapper to handle locking of mutexes
#   - will not lock forever
#   - when actual race condition occurs, allows easy debugging of who is locking who
#
class Lock:

    def __init__(
            self,
            mutex_names,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()

        self.__mutexes = {}
        for mtx_name in mutex_names:
            self.__mutexes[mtx_name] = {
                'mutex': threading.Lock(),
                'locked_by': None,
                'locked_since': datetime(year=2000, month=1, day=1),
            }
            self.logger.info('Created mutex lock for "' + str(mtx_name) + '"')
        return

    def is_locked(
            self,
            mutex,
    ):
        return self.__mutexes[mutex]['mutex'].locked()

    def acquire_mutexes(
            self,
            id,
            # list of mutex names, or plain string
            mutexes,
            try_count = 10,
            try_sleep_secs = 1.0,
            fake_race_condition_for_unit_test = False,
    ):
        mutexes = [mutexes] if type(mutexes) not in [list, tuple] else mutexes

        # if any exception occurs midway, we release all locked mutexes
        mutexes_locked_in_progress = []

        for mtx_name in mutexes:
            mtx = self.__mutexes[mtx_name]['mutex']
            fail_count = 0
            while True:
                fail_count += 1
                if fail_count >= try_count:
                    # release everything locked in this loop before
                    self.release_mutexes(mutexes=mutexes_locked_in_progress)
                    info = '"' + str(id) + '" wait for mutex "' + str(mtx_name) + '", tried ' + str(fail_count) + 'x.'
                    raise Exception('FAILED TO ACQUIRE mutex ' + str(info))

                if mtx.locked():
                    locked_by = self.__mutexes[mtx_name]['locked_by']
                    locked_since = self.__mutexes[mtx_name]['locked_since']
                    try:
                        tdiff = datetime.now() - locked_since
                    except Exception as ex_race_condition:
                        self.logger.info(
                            'Possible race condition, mutex must have been released by another thread while in this ' +
                            'loop setting "locked_since" to ' + str(locked_since) + ': ' + str(ex_race_condition)
                        )
                        time.sleep(try_sleep_secs)
                        continue
                    tdiff = tdiff.days*86400 + tdiff.seconds + tdiff.microseconds/1000000

                    info = '"' + str(id) + '" waiting for mutex "' + str(mtx_name) + '", tried ' + str(fail_count) \
                        + ' times, locked by "' + str(locked_by) + '", locked since "' + str(locked_since) \
                        + '" for ' + str(tdiff) + ' secs.'

                    self.logger.info(info + ' Trying again in ' + str(try_sleep_secs) + ' secs..')
                    time.sleep(try_sleep_secs)
                    continue
                else:
                    if fake_race_condition_for_unit_test:
                        mtx.acquire()
                    # timeout here only to detect race condition, since we are already sure mutex is free
                    res = mtx.acquire(timeout=1)
                    self.logger.debug('Mutex acquired or not = ' + str(res))
                    if not res:
                        # continue trying
                        time.sleep(try_sleep_secs)
                        continue
                    else:
                        mutexes_locked_in_progress.append(mtx_name)

                    self.__mutexes[mtx_name]['locked_by'] = id
                    self.__mutexes[mtx_name]['locked_since'] = datetime.now()

                    self.logger.info(
                        '"' + str(id) + '" acquired mutex "' + str(mtx_name) + '" successfully'
                    )
                    break
        return

    # When releasing a list of mutexes, make sure try/except so that none will be locked forever
    def release_mutexes(
            self,
            # list of mutex names, or plain string
            mutexes,
    ):
        mutexes = [mutexes] if type(mutexes) not in [list, tuple] else mutexes

        for mtx_name in mutexes:
            mtx = self.__mutexes[mtx_name]['mutex']
            try:
                if mtx.locked():
                    id_holder = self.__mutexes[mtx_name]['locked_by']
                    mtx.release()
                    self.__mutexes[mtx_name]['locked_by'] = None
                    self.__mutexes[mtx_name]['locked_since'] = None
                    self.logger.info(
                        'Mutex name "' + str(mtx_name) + '" released by "' + str(id_holder)
                        + '" successfully ' + str(mtx)
                    )
                else:
                    self.logger.debug('Mutex is not locked, not releasing ' + str(mtx))
            except Exception as ex:
                self.logger.error('Error releasing mutex ' + str(mtx) + ': ' + str(ex))


class LockUnitTest:
    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        self.lock1 = 'l1'
        self.lock2 = 'l2'
        self.lock_manager = Lock(logger=self.logger, mutex_names=[self.lock1, self.lock2])

        self.race_threads = []
        for i in range(10):
            thrd = threading.Thread(
                target = self.thread_for_test,
                args = [i, self.lock1],
            )
            self.race_threads.append(thrd)

        return

    def thread_for_test(
            self,
            thread_id,
            lock_name,
    ):
        sleep_secs = np.random.randint(low=5,high=10) / 10
        # Only thread id number 8 gets the lock
        sleep_secs = 0. if thread_id == 8 else sleep_secs
        time.sleep(sleep_secs)

        have_lock = False
        try:
            self.lock_manager.acquire_mutexes(
                id = thread_id,
                mutexes = lock_name,
                try_count = 2,
                try_sleep_secs = 0.5
            )
            have_lock = True
        except Exception as ex:
            self.logger.info('Thread for test exception ' + str(ex))

        test_res = have_lock==True if thread_id == 8 else have_lock==False
        assert test_res, 'Thread id "' + str(thread_id) + '", got lock "' + str(have_lock) + '"'
        return

    def race_to_get_lock(
            self,
    ):
        self.lock_manager.release_mutexes(mutexes=[self.lock1, self.lock2])
        for i in range(len(self.race_threads)):
            self.logger.info('Starting race thread #' + str(i))
            self.race_threads[i].start()
        return

    def test(self):
        lock1 = self.lock1
        lock2 = self.lock2
        obj = self.lock_manager

        obj.acquire_mutexes(id='1', mutexes=lock1, try_count=2, try_sleep_secs=0.5)
        assert obj.is_locked(mutex=lock1), 'Lock "' + str(lock1) + '" locked'

        # Test unable to obtain lock
        exc_msg = ''
        try:
            obj.acquire_mutexes(id='1', mutexes=lock1, try_count=2, try_sleep_secs=0.5)
        except Exception as ex:
            self.logger.info('Occured exception: ' + str(ex))
            exc_msg = str(ex)
        assert re.match(pattern='FAILED TO ACQUIRE mutex.*', string=exc_msg), \
            'Expect exception when cannot obtain lock1, but got exception message: ' + str(exc_msg)

        # Test unable to obtain lock by faking race condition
        obj.release_mutexes(mutexes=lock1)
        assert not obj.is_locked(mutex=lock1), 'Unlocked lock1'
        exc_msg = ''
        try:
            obj.acquire_mutexes(
                id='1', mutexes=lock1, try_count=2, try_sleep_secs=0.5,
                fake_race_condition_for_unit_test=True,
            )
        except Exception as ex:
            self.logger.info('Occured exception: ' + str(ex))
            exc_msg = str(ex)
        assert re.match(pattern='.*FAILED TO ACQUIRE mutex .*', string=exc_msg), \
            'Expect exception race condition lock1, but got exception message: ' + str(exc_msg)

        # Test able to obtain lock again
        obj.release_mutexes(mutexes=lock1)
        assert not obj.is_locked(mutex=lock1), 'Unlocked lock1'
        obj.acquire_mutexes(id='1', mutexes=[lock1, lock2], try_count=2, try_sleep_secs=0.5)
        assert obj.is_locked(mutex=lock1)
        assert obj.is_locked(mutex=lock2)

        print('Start real race to get lock')
        self.race_to_get_lock()

        for i in range(len(self.race_threads)):
            self.race_threads[i].join()

        print('ALL TESTS PASSED OK')
        return


if __name__ == '__main__':
    LockUnitTest(
        logger = Logging.get_default_logger(log_level=logging.INFO, propagate=False),
    ).test()
    exit(0)



# Fit
from fitxf.math.fit.cluster.ClusterUnitTest import Cluster, ClusterUnitTest
from fitxf.math.fit.cluster.ClusterCosineUT import ClusterCosine, ClusterCosineUnitTest
from fitxf.math.fit.transform.FitXformInterface import FitXformInterface
from fitxf.math.fit.transform.FitXformUnitTest import FitXformCluster, FitXformClusterCosine, FitXformPca
from fitxf.math.fit.transform.FitXformUnitTest import FitXformUnitTest
from fitxf.math.fit.utils.FitUtilsUnitTest import FitUtils, FitUtilsUt
from fitxf.math.fit.utils.TensorUtils import TensorUtils, TensorUtilsUnitTest
from fitxf.math.fit.utils.MathUtils import MathUtils
from fitxf.math.fit.utils.PatternSearch import PatternSearch
from fitxf.math.graph.GraphUtils import GraphUtils

# Cluster
from fitxf.math.fit.cluster.ClusterUnitTest import Cluster, ClusterUnitTest
from fitxf.math.fit.cluster.ClusterCosineUT import ClusterCosine, ClusterCosineUnitTest

# Algo
from fitxf.math.algo.encoding.Base64 import Base64, Base64UnitTest

# VecDB
from fitxf.math.datasource.DatastoreInterface import DatastoreInterface, DbParams, DatastoreInterfaceUnitTest
from fitxf.math.datasource.Datastore import Datastore
from fitxf.math.datasource.DbLang import DbLang
from fitxf.math.datasource.vecdb.metadata.MetadataInterface import MetadataInterface
from fitxf.math.datasource.vecdb.model.ModelDbInterface import ModelDbInterface
from fitxf.math.datasource.vecdb.model.ModelEncoderInterface import ModelEncoderInterface
from fitxf.math.datasource.vecdb.model.ModelInterface import ModelInterface
from fitxf.math.datasource.vecdb.model.ModelFitXForm import ModelFitTransform



import time
import logging
import os
import numpy as np
from datetime import datetime
from fitxf.math.datasource.DatastoreInterface import DbParams
from fitxf import FitXformInterface
from fitxf.math.lang.encode.LangModelPt import LangModelPt
from fitxf.math.lang.encode.ImgPt import ImgPt
from fitxf.math.datasource.vecdb.model.ModelDbInterface import ModelDbInterface
from fitxf.math.datasource.vecdb.model.ModelInterface import ModelInterface, ModelEncoderInterface
from fitxf.math.datasource.vecdb.metadata.MetadataInterface import MetadataInterface
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Env import Env


class ModelFitTransform(ModelInterface):

    def __init__(
            self,
            user_id: str,
            # e.g. {'text': modeltxt, 'image': modelimg, ...}
            llm_model: dict[str, ModelEncoderInterface],
            model_db_class: type(ModelDbInterface),
            model_metadata_class: type(MetadataInterface),
            col_content: str,
            col_content_type: str,
            col_label_user: str,
            col_label_std: str,
            col_embedding: str,
            feature_len: int,
            numpy_to_b64_for_db: bool,
            fit_xform_model: FitXformInterface,
            cache_tensor_to_file: bool,
            file_temp_dir: str,
            # turn off any fitting or transform if False
            in_plain_or_test_mode: bool,
            # allowed values: "np", "torch"
            return_tensors: str = 'np',
            enable_bg_thread_for_training: bool = False,
            logger = None,
    ):
        super().__init__(
            user_id = user_id,
            llm_model = llm_model,
            model_db_class = model_db_class,
            model_metadata_class = model_metadata_class,
            col_content = col_content,
            col_content_type = col_content_type,
            col_label_user = col_label_user,
            col_label_std = col_label_std,
            col_embedding = col_embedding,
            feature_len = feature_len,
            numpy_to_b64_for_db = numpy_to_b64_for_db,
            fit_xform_model = fit_xform_model,
            cache_tensor_to_file = cache_tensor_to_file,
            file_temp_dir = file_temp_dir,
            in_plain_or_test_mode = in_plain_or_test_mode,
            return_tensors = return_tensors,
            enable_bg_thread_for_training = enable_bg_thread_for_training,
            logger = logger,
        )

        # 1st time load data
        self.init_data_model(
            max_tries = 1,
            background = False,
        )
        self.check_model_consistency_with_prev()

        if self.enable_bg_thread_for_training:
            self.bg_thread.start()
        return

    def stop_threads(self):
        if not self.signal_stop_bg_thread:
            self.signal_stop_bg_thread = True
            if self.enable_bg_thread_for_training:
                self.bg_thread.join()
            self.logger.info('All threads successfully stopped')
        return

    def run_bg_thread(
            self,
    ):
        self.logger.info('Model thread started...')
        while True:
            time.sleep(self.bg_thread_sleep_secs)
            if self.signal_stop_bg_thread:
                self.logger.warning('Exiting model thread, stop signal received.')
                break
            if not self.is_need_sync_db():
                continue
            try:
                self.update_model(
                    test_mode = self.in_plain_or_test_mode,
                )
            except Exception as ex:
                self.logger.error('Error updating model compression from scheduled job: ' + str(ex))

    def init_data_model(
            self,
            max_tries = 1,
            background = False,
    ):
        # During initialization, we cannot throw exception. Fail means will depend on primary user cache already.
        required_mutexes = [self.mutex_name_model]
        try:
            self.lock_mutexes.acquire_mutexes(
                id = 'init_data_model',
                mutexes = required_mutexes,
            )
            self.__reset_data_model()
            # let Exceptions throw
        finally:
            self.lock_mutexes.release_mutexes(mutexes=required_mutexes)

        # get previous model data
        mtd_row = self.vec_db_metadata.get_metadata(identifier='model')
        self.logger.info(
            'Previous model type "' + str(type(mtd_row)) + '" info from metadata: '
            + str(mtd_row)
        )
        if not self.base64_encoder.is_base_64_string(mtd_row):
            self.logger.info(
                'Invalid model base64 string "' + str(mtd_row) + '", proceed to update model..'
            )
            self.update_model(
                test_mode = self.in_plain_or_test_mode,
            )
        else:
            try:
                self.lock_mutexes.acquire_mutexes(
                    id = 'init_data_model',
                    mutexes = required_mutexes,
                )
                model_b64json_str = mtd_row[MetadataInterface.COL_METADATA_VALUE]
                self.fit_xform_model.load_model_from_b64json(
                    model_b64json = model_b64json_str,
                )
                self.text_labels_standardized = np.array(self.fit_xform_model.X_labels)
                self.last_sync_time_with_underlying_db = self.vec_db_metadata.get_datetime_from_timestamp(
                    timestamp = mtd_row[MetadataInterface.COL_METADATA_TIMESTAMP]
                )
                self.logger.info(
                    'Loaded model ' + str(self.fit_xform_model.__class__)
                    + ' from saved model with timestamp ' + str(self.last_sync_time_with_underlying_db)
                    + ' metadata: ' + str(model_b64json_str)
                )
            finally:
                self.lock_mutexes.release_mutexes(mutexes=required_mutexes)
        return

    def load_data_model(
            self,
            max_tries = 1,
            background = False,
    ):
        return self.init_data_model(
            max_tries = max_tries,
            background = background,
        )

    def __reset_data_model(
            self,
    ):
        #
        # Default "built-in filesystem" model parameters required for inference
        #
        # In truth these maps are not required for this simple dense model, but we keep it anyway
        # since all other proper math models (e.g. NN) will always need mapping labels to numbers.
        self.map_lbl_to_idx = {}
        self.map_idx_to_lbl = {}
        self.text_labels_standardized = np.array([])
        self.data_records = []
        return

    def update_model(
            self,
            force_update = False,
            test_mode = False,
    ):
        if not force_update:
            if not self.is_need_sync_db():
                self.logger.warning('No longer required to update model...')
                return False

        self.logger.info('Model updating, test mode ' + str(test_mode) + '.')

        # Lock also underlying DB mutex because our metadata is also stored there
        required_mutexes = [self.mutex_name_model, self.mutex_name_underlying_db]
        try:
            self.lock_mutexes.acquire_mutexes(
                id = 'update_model',
                mutexes = required_mutexes,
            )
            db_records = self.model_db.load_data(max_attemps=1)

            self.data_records = [
                {
                    k: v for k, v in r.items()
                    if k in [self.col_content, self.col_label_standardized, self.col_label_user, self.col_embedding]
                }
                for r in db_records
            ]

            text_encoded = self.get_text_encoding_from_db_records(db_records=self.data_records)
            text_labels_user = [r[self.col_label_user] for r in self.data_records]
            self.text_labels_standardized = np.array([r[self.col_label_standardized] for r in self.data_records])

            unique_labels = len(np.unique(np.array(text_labels_user)))

            # not more than 3 per label, or not more than half the data length
            n_cluster = min(unique_labels * 3, int(len(text_labels_user)/2))
            self.logger.info(
                'Empirical estimation of total clusters ' + str(n_cluster) + ' from unique labels ' + str(unique_labels)
                + ', length of data ' + str(len(text_labels_user))
            )

            self.logger.info('Fitting to labels of user: ' + str(self.text_labels_standardized))
            if len(text_encoded) > 0:
                res = self.fit_xform_model.fine_tune(
                    X = text_encoded,
                    X_labels = text_labels_user,
                    X_full_records = self.data_records,
                    n_components = n_cluster,
                    test_mode = test_mode,
                )
                self.logger.debug(
                    'Fit to n cluster = ' + str(n_cluster) + ' result ' + str(res) + ', ' + str(text_labels_user)
                )

                model_save_b64json_string = self.fit_xform_model.model_to_b64json(
                    numpy_to_base64_str = True,
                    dump_to_b64json_str = True,
                )
                self.logger.debug(
                    'For user id "' + str(self.user_id) + '" model save json string: ' + str(model_save_b64json_string)
                )
                self.update_metadata_model_updated(
                    model_save_b64json_string = model_save_b64json_string,
                )
            else:
                # Update model metadata so that it will signal that we are done with loading data
                # Otherwise this function will be called non-stop
                self.update_metadata_model_updated(
                    model_save_b64json_string = None,
                )
                self.logger.info('No data or dont exist yet for "' + str(self.user_id) + '", nothing to fit.')

            mtd_row = self.vec_db_metadata.get_metadata(
                identifier = 'lastUpdateTimeDb',
            )
            if mtd_row is None:
                self.logger.warning('Last DB data update time from metadata returned None')
                self.last_sync_time_with_underlying_db = self.OLD_DATETIME
            else:
                self.last_sync_time_with_underlying_db = datetime.strptime(
                    mtd_row[MetadataInterface.COL_METADATA_VALUE], MetadataInterface.DATETIME_FORMAT
                )
            self.logger.info(
                'DB params: ' + str(self.model_db.get_db_params().get_db_info())
                + ', encode embedding to base 64 = ' + str(self.numpy_to_b64_for_db)
                + ', updated last sync time DB to "' + str(self.last_sync_time_with_underlying_db) + '"'
            )
            return True
        finally:
            self.lock_mutexes.release_mutexes(mutexes=required_mutexes)

    # Not necessarily faster, but will reduce RAM footprint
    def predict(
            self,
            text_list_or_embeddings,
            content_type: str = ModelInterface.TYPE_TEXT,
            # can be cluster numbers to zoom into
            X_embeddings_local_space: np.ndarray = None,
            labels_local_space: list = None,
            top_k = 5,
            # Instead of just returning the user labels, return full record. Applicable to some models only
            return_full_record = False,
    ):
        if self.is_need_sync_db():
            self.logger.info(
                'Model "' + str(self.fit_xform_model.__class__) + '" need update before prediction: '
                + str(text_list_or_embeddings)
            )
            self.update_model(
                test_mode = self.in_plain_or_test_mode,
            )
        else:
            self.logger.info(
                'Model "' + str(self.fit_xform_model.__class__) + '" dont need update before prediction: '
                + str(text_list_or_embeddings)
            )

        txt_lm = self.convert_to_embeddings_if_necessary(
            text_list_or_embeddings = text_list_or_embeddings,
            content_type = content_type,
        )
        # txt_lm = self.extend_feature_len(x=txt_lm)

        #
        # There are 2 possible approaches, after obtaining the PCA segment numbers & relevant reference vectors:
        #    1. Transform input vector to PCA transform, then compare with the reference PCA transforms
        #    2. Do not transform, use original vectors to compare. For now, we use this approach to skip the step
        #       of transforming the input vector.
        #

        pred_labels_std_or_full_records, pred_probs = self.fit_xform_model.predict(
            X = txt_lm,
            X_search_local_space = X_embeddings_local_space,
            labels_search_local_space = labels_local_space,
            top_k = top_k,
            return_full_record = return_full_record,
        )

        return pred_labels_std_or_full_records, pred_probs

    def atomic_delete_add(
            self,
            delete_key: str,
            # list of dicts
            records: list[dict],
    ):
        assert len(records) > 0, 'No records to train'
        self.logger.info('Add records of length ' + str(len(records)))

        cont_types = list(np.unique([r[self.col_content_type] for r in records]))
        if len(cont_types) > 1:
            encodings = []
            for i, row in enumerate(records):
                row_encode = self.calc_embedding(
                    content_list = [row[self.col_content]],
                    content_type = row[self.col_content_type],
                )
                # row_encode = self.extend_feature_len(x=row_encode)
                encodings.append(row_encode)
            content_encoding = np.vstack(encodings)
        else:
            content_encoding = self.calc_embedding(
                content_list = [r[self.col_content] for r in records],
                content_type = cont_types[0],
            )
            # content_encoding = self.extend_feature_len(x=content_encoding)
        self.logger.info(
            'Content of types ' + str(cont_types) + ' encoded using lm model "' + str(self.llm_model) + '" with shape '
            + str(content_encoding.shape if self.return_tensors == 'np' else content_encoding.size())
        )
        # txt_encoding = self.calc_embedding(
        #     content_list = [r[self.col_content] for r in records],
        # )
        self.logger.info(
            'Text encoded using lm model "' + str(self.llm_model) + '" with shape '
            + str(content_encoding.shape if self.return_tensors == 'np' else content_encoding.size())
        )

        required_mutexes = [self.mutex_name_underlying_db]
        try:
            self.lock_mutexes.acquire_mutexes(
                id = 'delete',
                mutexes = required_mutexes,
            )
            records_with_embedding_and_labelstd = self.update_label_maps_from_new_recs__(
                records = records,
                text_encoding_tensor = content_encoding,
            )
            mps = [{delete_key: r[delete_key]} for r in records_with_embedding_and_labelstd]
            try:
                self.delete_records_from_underlying_db__(
                    match_phrases = mps,
                )
            except Exception as ex_del:
                self.logger.info('Ignore delete error before add using match phrases ' + str(mps) + ': ' + str(ex_del))
                pass
            self.add_records_to_underlying_db__(
                records_with_embedding_and_labelstd = records_with_embedding_and_labelstd,
            )
            self.update_metadata_db_data_updated()
        finally:
            self.lock_mutexes.release_mutexes(mutexes=required_mutexes)

        # if not self.enable_bg_thread_for_training:
        #     self.update_model()
        #     is_updated = self.update_model()
        #     self.logger.info(
        #         'Model is updated = "' + str(is_updated) + '" after add records '
        #         + str([r[self.col_content] for r in records])
        #     )
        return

    def add(
            self,
            # list of dicts
            records: list,
    ):
        assert len(records) > 0, 'No records to train'
        self.logger.info('Add records of length ' + str(len(records)))
        cont_types = list(np.unique([r[self.col_content_type] for r in records]))
        if len(cont_types) > 1:
            encodings = []
            for i, row in enumerate(records):
                row_encode = self.calc_embedding(
                    content_list = [row[self.col_content]],
                    content_type = row[self.col_content_type],
                )
                # row_encode = self.extend_feature_len(x=row_encode)
                encodings.append(row_encode)
            content_encoding = np.vstack(encodings)
        else:
            content_encoding = self.calc_embedding(
                content_list = [r[self.col_content] for r in records],
                content_type = cont_types[0],
            )
            # content_encoding = self.extend_feature_len(x=content_encoding)
        self.logger.info(
            'Content of types ' + str(cont_types) + ' encoded using lm model "' + str(self.llm_model) + '" with shape '
            + str(content_encoding.shape if self.return_tensors == 'np' else content_encoding.size())
        )

        required_mutexes = [self.mutex_name_underlying_db]
        try:
            self.lock_mutexes.acquire_mutexes(
                id = 'add',
                mutexes = required_mutexes,
            )
            records_with_embedding_and_labelstd = self.update_label_maps_from_new_recs__(
                records = records,
                text_encoding_tensor = content_encoding,
            )
            self.add_records_to_underlying_db__(
                records_with_embedding_and_labelstd = records_with_embedding_and_labelstd,
            )
            self.update_metadata_db_data_updated()
        finally:
            self.lock_mutexes.release_mutexes(mutexes=required_mutexes)

        # if not self.enable_bg_thread_for_training:
        #     is_updated = self.update_model()
        #     self.logger.info(
        #         'Model is updated = "' + str(is_updated) + '" after add records '
        #         + str([r[self.col_content] for r in records])
        #     )
        return

    def delete(
            self,
            match_phrases,
    ):
        required_mutexes = [self.mutex_name_underlying_db]
        try:
            self.lock_mutexes.acquire_mutexes(
                id = 'delete',
                mutexes = required_mutexes,
            )
            self.delete_records_from_underlying_db__(
                match_phrases = match_phrases,
            )
            self.update_metadata_db_data_updated()
        finally:
            self.lock_mutexes.release_mutexes(mutexes=required_mutexes)

        # if not self.enable_bg_thread_for_training:
        #     self.update_model()
        #     is_updated = self.update_model()
        #     self.logger.info(
        #         'Model is updated = "' + str(is_updated) + '" after delete records ' + str(match_phrases)
        #     )
        return


if __name__ == '__main__':
    from fitxf import FitXformClusterCosine, FitXformCluster, FitXformPca
    from nwae.math.datasource.vecdb.model.ModelDb import ModelDb
    from nwae.math.datasource.vecdb.metadata.Metadata import ModelMetadata
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    er = Env()
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/.env.fitxf.math.ut')
    user_id = 'test_modelfitxf'
    db_prms = DbParams.get_db_params_from_envvars(
        identifier = 'test_modelfitxf',
        db_create_tbl_sql = None,
        db_table = user_id,
    )
    model_fit = ModelFitTransform(
        user_id = user_id,
        llm_model = {
            ModelInterface.TYPE_TEXT: LangModelPt(
                # model_name = 'intfloat/multilingual-e5-small',
                # model_name = 'sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2',
                cache_folder = er.MODELS_PRETRAINED_DIR,
                logger = lgr,
            ),
            ModelInterface.TYPE_IMG: ImgPt(
                cache_folder = er.MODELS_PRETRAINED_DIR,
                logger = lgr,
            )
        },
        model_db_class = ModelDb,
        model_metadata_class = ModelMetadata,
        col_content = 'c',
        col_content_type = 't',
        col_label_user = 'l',
        col_label_std = '__l',
        col_embedding = 'e',
        feature_len = 384,
        fit_xform_model = FitXformCluster(logger=lgr),
        numpy_to_b64_for_db = True,
        file_temp_dir = er.REPO_DIR + '/tmp',
        cache_tensor_to_file = False,
        enable_bg_thread_for_training = False,
        logger = lgr,
    )
    recs_test = [
        {'l': 'fruits', 't': 'text', 'c': 'mango'}, {'l': 'fruits', 't': 'text', 'c': 'pineapple'},
        {'l': 'fruits',  't': 'text', 'c': 'apple'}, {'l': 'fruits',  't': 'text', 'c': 'orange'},
        {'l': 'beer',  't': 'text', 'c': 'Мюнхенский хеллес или Хелль '},
        {'l': 'beer', 't': 'text', 'c': 'Грузинский Черный Лев'},
        {'l': 'plov', 't': 'image',
         'c': 'https://img.freepik.com/premium-photo/shakh-plov-cooked-rice-dish-with-raisins-beautiful-plate-islamic-arabic-food_1279579-5074.jpg?w=1800'},
        {'l': 'plov', 't': 'image',
         'c': 'https://img.freepik.com/premium-psd/tasty-fried-vegetable-rice-plate-isolated-transparent-background_927015-3126.jpg?w=1480'},
    ]
    model_fit.atomic_delete_add(records=recs_test, delete_key='c')

    for test in [
        'давай выпем',
        'https://www.alyonascooking.com/wp-content/uploads/2019/05/plov-recipe-6.jpg',
    ]:
        res = model_fit.predict(text_list_or_embeddings=[test], top_k=2)
        print(res)
    exit(0)



import logging
import numpy as np
import pandas as pd
from fitxf.math.utils.Logging import Logging


class Metrics:

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    """
    In some cases, we may have labelled data which we subject to clustering (e.g. using text embedding)
    For example:
    
           label                                               text  cluster_number
            food                                   bread and butter               3
            food                                     fish and chips               1
            food            sausages, scrambled eggs or hash browns               1
            tech                              computers and laptops               8
            tech                                   code programming               4
            tech                                    8 bits one byte               4
          sports                                 Tennis grass court               9
          sports                                   Soccer world cup               9
          sports                     50m freestyle record under 21s               7
        medicine                    Diagnosis and treatment options               7
        medicine  Changing lifestyle habits over surgery & presc...               2
        medicine        Genomic basis for RNA alterations in cancer               6
    
    Then we want to find the "purity" of the cluster numbers, whether or not each label is assigned.
    Min purity = 1.0,
    Max purity = total labels
    """
    def get_label_cluster_purity(
            self,
            point_cluster_numbers: list,
            point_labels: list,
    ):
        n_clusters = len(np.unique(point_cluster_numbers))
        assert np.max(point_cluster_numbers) == n_clusters-1
        assert np.min(point_cluster_numbers) == 0

        self.logger.info('Total unique cluster numbers ' + str(n_clusters))
        df = pd.DataFrame({'label': point_labels, 'cluster_number': point_cluster_numbers})
        # We want to extract something like this: labels --> point clusters
        #   food [3, 1, 1, 1, 1, 1, 1, 1, 3, 3, 3, 1, 1, 1]
        #   tech [8, 4, 4, 8, 4, 4, 4, 4, 4, 4, 8, 4, 4, 4]
        #   sports [9, 9, 7, 9, 9, 9, 9, 7, 9, 9, 7, 7, 9, 5]
        #   medicine [7, 2, 6, 7, 2, 5, 4, 5, 5, 5, 4, 5, 0]
        #   genetics [6, 6, 6, 6, 2, 2, 6, 2, 2, 6, 2, 2, 6, 2]
        label_to_cluster_numbers = {lbl: [] for lbl in pd.unique(df['label'])}
        # Count instead (how many in each cluster number)
        #   food [ 0 10  0  4  0  0  0  0  0  0]
        #   tech [ 0  0  0  0 11  0  0  0  3  0]
        #   sports [0 0 0 0 0 1 0 4 0 9]
        #   medicine [1 0 2 0 2 5 1 2 0 0]
        #   genetics [0 0 7 0 0 0 7 0 0 0]
        label_to_cluster_counts = {lbl: np.zeros(n_clusters, dtype=int) for lbl in pd.unique(df['label'])}
        # Same as above, but not by label, just globally
        label_to_cluster_counts_global = np.zeros(n_clusters)
        self.logger.info('Global counts: ' + str(label_to_cluster_counts_global))
        # cluster_freq_all = np.array(cluster_numbers)
        for i, r in enumerate(df.to_records()):
            c_no = r['cluster_number']
            label_to_cluster_numbers[r['label']].append(c_no)
            label_to_cluster_counts[r['label']][c_no] += 1
            label_to_cluster_counts_global[c_no] += 1
        label_to_cluster_purity = {k: v for k, v in label_to_cluster_counts.items()}
        for lbl, val in label_to_cluster_counts.items():
            # normalize such that sum is 1
            val_prob = val / np.sum(val)
            # Dividing by the global value will give us the "true" value, meaning that if for "food", it
            # is in cluster 1 ten times, and globally there are also 10 counts, then it has a pure 1.0
            # concentration.
            # So for example the data [[0,0,0,0], [1,1,1,1] will have the ownership
            # [[1.0,1.0], [1.0,1.0]] or [1,1] after multiply with prob
            # wheres the data [[0,0,1,1], [1,1,1,1]] or freq [[2,2], [0,4]]
            # will have ownership [[1,0.33],[0,0.66]] or normalized by prob to [[0.5+0.166],[0+0.666]]
            # with the 1st vector having higher "purity" of 0.
            label_to_cluster_purity[lbl] = np.sum( (val / label_to_cluster_counts_global) * val_prob )

        # [print(lbl, m) for lbl, m in label_to_cluster_numbers.items()]
        # [print(lbl, m) for lbl, m in label_to_cluster_counts.items()]
        # print('ownership', label_to_cluster_ownership)
        # print('global', label_to_cluster_counts_global)
        aggregated_purity = np.sum(np.array(list(label_to_cluster_purity.values()))) / len(label_to_cluster_purity)
        return {
            'label_purity': label_to_cluster_purity,
            'final_purity': aggregated_purity,
        }

    def map_cluster_labels_to_original_labels(
            self,
            point_cluster_numbers: list,
            point_labels: list,
    ):
        labels_unique = np.unique(point_labels).tolist()
        labels_unique.sort()
        map_index_to_labels = {i: lbl for i, lbl in enumerate(labels_unique)}
        map_labels_to_index = {v: k for k, v in map_index_to_labels.items()}
        n_labels_unique = len(labels_unique)
        # It is possible some cluster labels have no clusters, thus taking max number is more correct
        n_cno_unique = np.max(point_cluster_numbers) + 1

        self.logger.info('Total unique original labels ' + str(n_labels_unique))
        df = pd.DataFrame({'label': point_labels, 'cluster_number': point_cluster_numbers})
        self.logger.info('Label and cluster numbers: ' + str(df))
        # For example, given the following labels & cluster numbers
        #      ['a', 'a', 'a', 'b', 'b', 'b', 'c', 'c', 'c'],
        #      [  0,   0,   1,   0,   1,   1,   0,   2,   2],
        # We want to extract something like this: point clusters --> labels
        #      0: ['a', 'a', 'b', 'c']
        #      1: ['a', 'b', 'b']
        #      2: ['c', 'c']
        clusterno_to_labelori = {cno: [] for cno in pd.unique(df['cluster_number'])}
        # Count instead (how many in each label)
        #      {0: array([2, 1, 1]), 1: array([1, 2, 0]), 2: array([0, 0, 2])}
        clusterno_to_labelori_counts = {
            cno: np.zeros(n_labels_unique, dtype=int) for cno in pd.unique(df['cluster_number'])
        }
        # Same as above, but not by label, just globally
        #       [4. 3. 2.]
        # Meaning cluster #0 appeared in 4 places, cluster #1 appeared in 3 places, and cluster #2 in 2 places
        clusterno_to_labelori_counts_global = np.zeros(n_cno_unique)
        # cluster_freq_all = np.array(cluster_numbers)
        for i, r in enumerate(df.to_records()):
            c_no = r['cluster_number']
            lbl_ori = r['label']
            clusterno_to_labelori[c_no].append(lbl_ori)
            clusterno_to_labelori_counts[c_no][map_labels_to_index[lbl_ori]] += 1
            clusterno_to_labelori_counts_global[c_no] += 1
        self.logger.info('Cluster no to original labels: ' + str(clusterno_to_labelori))
        self.logger.info('Cluster no to original labels count: ' + str(clusterno_to_labelori_counts))
        self.logger.info('Cluster no to original labels count global: ' + str(clusterno_to_labelori_counts_global))

        clusterno_to_labelori_probs = {k: v/v.sum() for k, v in clusterno_to_labelori_counts.items()}
        self.logger.info('Cluster no to original labels indexes probabilities: ' + str(clusterno_to_labelori_probs))

        # Change to more convenient form
        for k, np_prob in clusterno_to_labelori_probs.items():
            clusterno_to_labelori_probs[k] = {map_index_to_labels[i]: prob for i, prob in enumerate(np_prob)}
        self.logger.info('Cluster no to original labels probabilities: ' + str(clusterno_to_labelori_probs))

        # Probability map now looks like this
        #       {
        #          0: {'a': 0.5,   'b': 0.25,  'c': 0.25},
        #          1: {'a': 0.333, 'b': 0.666, 'c': 0.0 },
        #          2: {'a': 0.0,   'b': 0.0,   'c': 1.0 }
        #       }
        # So cluster #0 means 50% probability belongs to "a", 25% to "b", 25% to "c"
        # Cluster #2 means 33.33% "a", 66.66% "b" and 0% "c"
        # Cluster #3 means 100% sure is "c"

        # Sort a little for easier reading & easier processing when returning search results to applications
        # that can easily pick top 1 from index 0
        clusterno_to_labelori_probs_sorted = {}
        for cno, label_probs in clusterno_to_labelori_probs.items():
            df_k = pd.DataFrame({'label': label_probs.keys(), 'prob': label_probs.values()})
            df_k = df_k.sort_values(
                by = ['prob'],
                ascending = False,
            )
            clusterno_to_labelori_probs_sorted[cno] = {
                rec['label']: rec['prob'] for rec in df_k.to_dict(orient='records')
            }

        return clusterno_to_labelori_probs_sorted


if __name__ == '__main__':
    m = Metrics(logger=Logging.get_default_logger(log_level=logging.INFO, propagate=False))
    point_labels_cno = [
        # Max purity - all different
        [
            ['a', 'a', 'a', 'b', 'b', 'b', 'c', 'c', 'c'],
            [  0,   0,   0,   1,   1,   1,   2,   2,   2],
        ],
        # Min purity, all distributed
        [
            ['a', 'a', 'a', 'b', 'b', 'b', 'c', 'c', 'c'],
            [  0,   1,   2,   0,   1,   2,   0,   1,   2],
        ],
        # Some mixture
        [
            ['a', 'a', 'a', 'b', 'b', 'b', 'c', 'c', 'c'],
            [  0,   0,   1,   0,   1,   1,   0,   2,   2],
        ],
    ]
    for point_labels, point_cluster_numbers in point_labels_cno:
        purity = m.get_label_cluster_purity(
            point_cluster_numbers = point_cluster_numbers,
            point_labels = point_labels,
        )
        print(purity)

        map_cno_lbl = m.map_cluster_labels_to_original_labels(
            point_cluster_numbers = point_cluster_numbers,
            point_labels = point_labels,
        )
        print(map_cno_lbl)
    exit(0)



import logging
import os
import numpy as np
import pandas as pd
from io import StringIO
from fitxf.math.data.ut.LabelTextEmbed01 import DATA_LABEL_TEXT_EMBEDDING_01_TRAIN
from fitxf.math.algo.encoding.Base64 import Base64
from fitxf.math.fit.utils.TensorUtils import TensorUtils
from fitxf.math.fit.cluster.Metrics import Metrics as ClusterMetrics
from fitxf.math.fit.cluster.ClusterCosine import ClusterCosine
from fitxf.math.lang.encode.LangModelPtSingleton import LangModelPtSingleton, LangModelInterface
from fitxf.math.lang.encode.LangModelPt import LangModelPt
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Env import Env
from fitxf.math.utils.Pandas import Pandas


class ClusterCosineUnitTest:

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        self.tensor_utils = TensorUtils(logger=self.logger)
        self.cluster_cos = ClusterCosine(logger=self.logger)
        self.base64 = Base64(logger=self.logger)
        return

    def test_nlp(
            self,
    ):
        er = Env()
        sample_data_csv = er.NLP_DATASET_DIR + '/lang-model-test/data.csv'
        # Force to only use pre-calculated embedding
        # is_csv_exist = os.path.exists(sample_data_csv)
        is_csv_exist = False
        self.logger.info('Test data csv exist = ' + str(is_csv_exist) + '.')

        def get_lm() -> LangModelInterface:
            return LangModelPtSingleton.get_singleton(
                LmClass = LangModelPt,
                cache_folder = er.MODELS_PRETRAINED_DIR,
                logger = self.logger,
            )

        Pandas.increase_display()
        df = pd.read_csv(
            filepath_or_buffer = sample_data_csv if is_csv_exist else StringIO(DATA_LABEL_TEXT_EMBEDDING_01_TRAIN),
            sep = ',',
            index_col = False,
        )
        columns_keep = ['label', 'text'] if is_csv_exist else ['label', 'text', 'embedding']
        for c in columns_keep: assert c in df.columns
        df = df[columns_keep]
        df.dropna(inplace=True)
        # _, _, df[self.col_label_std] = FitUtils().map_labels_to_consecutive_numbers(lbl_list=list(df[self.col_label]))
        self.logger.info('Successfully read data of shape ' + str(df.shape))
        if is_csv_exist:
            self.logger.info('Not using precalculated embeddings, recalculating text embeddings..')
            embeddings = get_lm().encode(
                content_list = df['text'].tolist(),
                return_tensors = 'np',
            )
            df['embedding'] = [
                self.base64.encode_numpy_array_to_base64_string(x=row, data_type=np.float64)
                for row in embeddings
            ]
            #df.to_csv('data_with_embedding.csv', index=False)
            #raise Exception('asdf')
        else:
            self.logger.info('Using precalculated embeddings, not calculating.')
            embeddings = np.array([
                self.base64.decode_base64_string_to_numpy_array(s64=s, data_type=np.float64)
                for s in df['embedding'].tolist()
            ])
        self.logger.info('Embeddings shape ' + str(embeddings.shape) + ', type "' + str(type(embeddings)) + '"')
        records = df.to_dict('records')
        # for i, r in enumerate(records):
        #     r['embedding'] = embeddings[i]
        #     print(i, r)

        n_unique_labels = 2 * len(pd.unique(df['label']))
        res = self.cluster_cos.kmeans(
            x = embeddings,
            n_centers = n_unique_labels,
            x_labels = df['label'].tolist(),
            km_iters = 100,
        )
        cluster_centers = np.array(res['cluster_centers'])
        point_cluster_numbers = res['cluster_labels']
        prob_cluster_no_to_labels = res['cluster_label_to_original_labels']
        self.logger.info('Probs cluster no to labels: ' + str(prob_cluster_no_to_labels))

        cluster_metrics = ClusterMetrics(logger=self.logger)
        purity = cluster_metrics.get_label_cluster_purity(
            point_labels = df['label'].tolist(),
            point_cluster_numbers = point_cluster_numbers,
        )
        label_purity, agg_purity = purity['label_purity'], purity['final_purity']
        self.logger.info('Label purity: ' + str(label_purity) + ', aggregated purity ' + str(agg_purity))
        min_agg_purity = 1.0 / len(label_purity)
        thr = min_agg_purity + 0.5 * (1 - min_agg_purity)
        self.logger.info('Aggregate purity ' + str(agg_purity) + ', threshold ' + str(thr))
        assert agg_purity > thr, \
            'Aggregated purity of cluster numbers ' + str(agg_purity) + ' not good enough ' + str(label_purity)

        # At this point, the application can represent each embedding with just the centroids
        # (thus cluster number as index)
        predicted_clusters, _ = self.tensor_utils.similarity_cosine(
            x = embeddings,
            ref = cluster_centers,
            return_tensors = 'np',
        )
        score_total = 0.
        for i, r in enumerate(records):
            c_no = point_cluster_numbers[i]
            r['cluster_number'] = c_no
            r['cluster_predicted_1'] = predicted_clusters[i, 0]
            r['cluster_predicted_2'] = predicted_clusters[i, 1]
            # Get top few predictions
            top2_label_prob = {
                k: p for i, (k, p) in enumerate(prob_cluster_no_to_labels[predicted_clusters[i, 0]].items()) if i<=1
            }
            r['pred1_label'] = list(top2_label_prob.keys())[0]
            r['pred1_prob'] = list(top2_label_prob.values())[0]
            r['pred2_label'] = list(top2_label_prob.keys())[1]
            r['pred2_prob'] = list(top2_label_prob.values())[1]
            score_total += min(1.0, 1*(r['pred1_label']==r['label']) + 0.5*(r['pred2_prob']==r['label']))

            c_pr = r['cluster_predicted_1']
            assert c_pr == c_no, \
                'For record #' + str(i) + ', text "' + str(r['text']) + '", cluster number ' \
                + str(c_no) + ' != cluster predicted ' + str(c_pr)

        score_avg = score_total / len(records)
        self.logger.info('Score average = ' + str(score_avg))
        thr_score = 0.88
        assert score_avg > thr_score, \
            'Average score in prediction ' + str(score_avg) + ' < ' + str(thr_score) \
            + ': ' + str(pd.DataFrame.from_records(records))
        print(pd.DataFrame.from_records(records))
        print('CLUSTER COSINE NLP TEST PASSED')
        return

    def test(self):
        self.test_basic()
        self.test_fine_tune()
        self.test_nlp()
        print('CLUSTER COSINE TESTS PASSED')
        return

    def test_basic(self):
        x = np.array([
            [1.0, 0.1, 0.1],
            [2.0, 0.2, -0.1],
            [-1.0, 1.0, 1.0],
            [-3.4, 2.0, 1.8],
        ])
        for n, exp_clusters, exp_centroids, exp_cluster_numbers in [
            (1, [[0, 1, 2, 3]],
             [[0.1556, 0.3092, 0.2604]],
             [0, 0, 0, 0]),
            (2, [[0, 1], [2, 3]],
             [[0.9919, 0.0991, 0.0246], [-0.6807, 0.5193, 0.4962]],
             [0, 0, 1, 1]),
            (3, [[0, 1], [2], [3]],
             [[0.9919, 0.0991, 0.0246], [-0.5773, 0.5773, 0.5773], [-0.7841, 0.4612, 0.4151]],
             [0, 0, 1, 2]),
            (4, [[0], [1], [2], [3]],
             [[0.9901, 0.09901, 0.09901], [0.9938, 0.0993, -0.0496], [-0.5773, 0.5773, 0.57735], [-0.7841, 0.4612, 0.4151]],
             [0, 1, 2, 3]),
        ]:
            res = self.cluster_cos.cluster_angle(
                x = x,
                n_clusters = n,
                max_iter = 10,
                start_min_dist_abs = 0.8,
            )
            clusters = res['clusters']
            centroids = res['centroids']
            cluster_numbers = res['cluster_numbers']
            print('------------------------------------------')
            print('clustering result', res)
            print('observed', clusters, 'expected', exp_clusters)
            assert str(clusters) == str(exp_clusters), \
                'For n=' + str(n) + ', observed clusters ' + str(clusters) + ' not expected ' + str(exp_clusters)
            diff_error = np.sum( ( np.array(centroids) - np.array(exp_centroids) ) ** 2 )
            assert diff_error < 0.0000001, \
                'Observed centroids ' + str(centroids) + ' not expected ' + str(exp_centroids)
            assert cluster_numbers.tolist() == exp_cluster_numbers, \
                    'Cluster numbers ' + str(cluster_numbers.tolist()) + ' not expected ' + str(exp_cluster_numbers)

            res = self.cluster_cos.kmeans(
                x = x,
                n_centers = n,
                km_iters = 10,
            )
            print('++++++++++++++++++++++++++++++++++++++++++')
            print('By kmeans: ' + str(res))
            print(
                'Centroid changed from ' + str(centroids) + ' to ' + str(res['cluster_centers'])
                + ' Clusters: ' + str(res['clusters'])
            )

        x = np.array([
            [1.0, 0.1, 0.1], [2.0, 0.2, -0.1],
            [-1.0, 1.0, 1.0], [-3.4, 2.0, 1.8],
            [-2.0, -2.3, -1.8], [-111, -100, -112],
        ])
        res = self.cluster_cos.kmeans_optimal(
            x = x,
        )
        assert len(res) == 1, 'Expect only 1 turning point but got ' + str(len(res)) + ': ' + str(res)
        # get first turning point
        cluster_last_turn_point = res[0]
        self.logger.info('Kmeans optimal at first turning point: ' + str(cluster_last_turn_point))
        optimal_clusters = cluster_last_turn_point['clusters']
        optimal_cluster_labels = cluster_last_turn_point['cluster_labels']
        assert len(optimal_clusters) == 3, 'No of optimal clusters ' + str(len(optimal_clusters)) + ' not expected 3'
        for i, j in [(0, 1), (2, 3), (4, 5)]:
            assert optimal_cluster_labels[i] == optimal_cluster_labels[j], \
                'Points ' + str((i,j)) + ' not in same cluster ' + str(optimal_cluster_labels)
        self.logger.info('Basic tests passed')
        return

    def test_fine_tune(self):
        x = np.array([
            [1.0, 0.1, 0.1], [2.0, 0.2, -0.1],
            [-1.0, 1.0, 1.0], [-3.4, 2.0, 1.8],
            [-2.0, -2.3, -1.8], [-111, -100, -112],
        ])
        res = self.cluster_cos.kmeans_optimal(
            x = x,
        )
        # get first turning point
        cluster_last_turn_point = res[0]
        self.logger.info('Kmeans optimal at first turning point: ' + str(cluster_last_turn_point))
        optimal_n = cluster_last_turn_point['n_centers']
        optimal_cluster_centers = cluster_last_turn_point['cluster_centers']

        #
        # Test fine-tuning
        #
        # add new point
        x_add = np.array([[1.1, 0.2, 0.1], [-2.1, -2.0, -1.9]])
        x_new = np.append(x, x_add, axis=0)
        res = self.cluster_cos.kmeans(
            x = x_new,
            n_centers = optimal_n,
            start_centers = optimal_cluster_centers,
            km_iters = 10,
        )
        original_centers = res['cluster_centers']
        original_center_lbls = res['cluster_labels']

        # Last added point cluster number must equal 1st one
        assert original_center_lbls[-2] == original_center_lbls[0], \
            'Last added point should belong to cluster of 1st 3 points but got labels ' + str(original_center_lbls)
        assert original_center_lbls[-1] == original_center_lbls[4], \
            'Last added point should belong to cluster of 1st 3 points but got labels ' + str(original_center_lbls)
        # This is same as above, test that original clusters retained
        for i, j in [(0, 1), (2, 3), (4, 5)]:
            assert original_center_lbls[i] == original_center_lbls[j], \
                'Points ' + str((i,j)) + ' not in same cluster ' + str(original_center_lbls)

        #
        # Test fine-tuning with new cluster center, make sure n remains the same after training
        #
        # add new point
        x_add = np.array([[1.1, 0.2, 0.1], [-2.1, -2.0, -1.9]])
        x_new = np.append(x, x_add, axis=0)
        res = self.cluster_cos.kmeans(
            x = x_new,
            n_centers = optimal_n + 1,
            start_centers = np.append(original_centers, np.array([[100., -100., 100.]]), axis=0),
            km_iters = 10,
        )
        n = res['n_centers']
        clusters = res['clusters']
        new_center_lbls = res['cluster_labels']
        for clstr in clusters:
            assert len(clstr) > 0, 'Empty cluster after fine-tuning: ' + str(clusters)

        # Last added point cluster number must equal 1st one
        self.logger.info(
            'After fine tuning add another center far away, n = ' + str(n) + ', cluster labels ' + str(new_center_lbls)
            + ', clusters ' + str(clusters)
        )
        assert n == optimal_n + 1


        #
        # Test fine-tuning with new cluster center, but force it to be empty by only allowing 1 iteration
        #
        # add new point
        x_add = np.array([[1.1, 0.2, 0.1], [-2.1, -2.0, -1.9]])
        x_new = np.append(x, x_add, axis=0)
        res = self.cluster_cos.kmeans(
            x = x_new,
            n_centers = optimal_n + 1,
            start_centers = np.append(original_centers, np.array([[100., -100., 100.]]), axis=0),
            km_iters = 1,
        )
        n = res['n_centers']
        clusters = res['clusters']
        new_center_lbls = res['cluster_labels']

        # Last added point cluster number must equal 1st one
        self.logger.info(
            'After fine tuning add another center far away, n = ' + str(n) + ', cluster labels ' + str(new_center_lbls)
            + ', clusters ' + str(clusters)
        )
        assert n == optimal_n + 1
        return


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    ClusterCosineUnitTest(
        logger = lgr,
    ).test()
    exit(0)



import logging
import math
import os
import json
import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from fitxf.math.fit.transform.FitXformInterface import FitXformInterface
from fitxf.math.utils.Lock import Lock
from fitxf.math.utils.Env import Env
from fitxf.math.utils.Logging import Logging


#
# Simple and easy to understand article on PCA
# https://ko.wikipedia.org/wiki/%EC%A3%BC%EC%84%B1%EB%B6%84_%EB%B6%84%EC%84%9D
#
# We build these on top of the PCA
#    - Divide the points using PCA components (positive or negative direction) into grids. k components form 2^k grids
#    - Measure grid density (how many points per grid)
#    - Measure distance error of estimated X from PCA & actual X
#    - Use both measures to find optimal number of k, based on measure value specified by user
#
class FitXformPca(FitXformInterface):

    def __init__(
            self,
            logger = None,
    ):
        super().__init__(
            logger = logger,
        )

        self.__mutex_model = 'model'
        self.__lock = Lock(
            mutex_names = [self.__mutex_model],
            logger = self.logger,
        )

        # Model parameters
        self.model_params_ready = False
        return

    def is_model_ready(self):
        return self.model_params_ready

    def __check_consistency(
            self,
            principal_components: np.ndarray,
    ):
        assert self.__lock.is_locked(mutex=self.__mutex_model)

        # Principal components must be of unit length
        tmp = principal_components * principal_components
        pc_lengths = np.sum(tmp, axis=-1) ** 0.5
        expected = np.ones(len(principal_components))
        diff = expected - pc_lengths
        diff = np.sum(diff * diff)
        assert diff < 0.0000000001, 'Expect principal components to be length 1 but got ' + str(pc_lengths)
        return

    def fit_optimal(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            target_grid_density = 2,
            # allowed values 'median', 'mean', 'min'
            measure = 'median',
            # Model dependent interpretation, or ignore if not relevant for specific model
            min_components = 1,
            max_components = 99999,
            # pass through mode
            test_mode = False,
    ) -> dict:
        try:
            self.__lock.acquire_mutexes(
                id = 'fit_optimal',
                mutexes = [self.__mutex_model],
            )
            return self.__fit_optimal(
                X = X,
                X_labels = X_labels,
                X_full_records = X_full_records,
                target_grid_density = target_grid_density,
                measure = measure,
                min_components = min_components,
                max_components = max_components,
            )
        finally:
            self.__lock.release_mutexes(mutexes=[self.__mutex_model])

    # Will dynamically look for optimal number of pca components, on the condition of the target grid density
    def __fit_optimal(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            target_grid_density = 2,
            # allowed values 'median', 'mean', 'min'
            measure = 'median',
            # Model dependent interpretation, or ignore if not relevant for specific model
            min_components = 1,
            max_components = 99999,
            # pass through mode
            test_mode = False,
    ) -> dict:
        assert target_grid_density > 0, 'Target grid density not valid ' + str(target_grid_density)
        x_dim = X.shape[-1]
        n_high = min(int(x_dim - 1), len(X)- 1)
        n_high = min(max_components, n_high)
        n_high_condition_satisfied = False
        n_low = max(1, min_components)
        n = n_low + math.ceil((n_high - n_low) / 2)

        pca_fit_optimal = None
        while True:
            self.logger.info('Fit PCA with n pca components = ' + str(n))
            pca_fit = self.__fit(
                X = X,
                X_labels = X_labels,
                X_full_records = X_full_records,
                n_components = n,
            )
            self.logger.info(
                'Fit PCA n=' + str(n) + ' returned keys ' + str(pca_fit.keys())
            )
            measure_value = self.grid_density_mean
            if measure_value == 'median':
                measure_value = np.median(self.grid_density)
            elif measure_value == 'min':
                measure_value = np.min(self.grid_density)

            info = 'n pca components = ' + str(n) + ', condition grid density "' + str(measure) \
                   + '" ' + str(measure_value) + ' >= ' + str(target_grid_density)\
                   + ', with distance mean error ' + str(self.distance_error_mean)

            if measure_value >= target_grid_density:
                n_low = n
                # Raise up number of components
                n += math.ceil((n_high - n) / 2)
                self.logger.info(
                    'SATISFIED ' + info + '. low=' + str(n_low) + ', high=' + str(n_high) + ', next n=' + str(n)
                )
                if n >= n_high:
                    # Only move back up if at n=n_high, condition is satisfied. Otherwise we have found our optimum
                    if (n_low < n_high) and n_high_condition_satisfied:
                        n_low = n_high
                        n = n_high
                    else:
                        self.logger.info(
                            'DONE at ' + info + '. low=' + str(n_low) + ', high=' + str(n_high)
                            + ', final grid density "' + str(measure) + '"=' + str(measure_value)
                            # + ', details: ' + str(self.grid_density)
                        )
                        pca_fit_optimal = pca_fit
                        break
            else:
                n_high = n
                n_high_condition_satisfied = False
                # Reduce by half
                n -= math.ceil((n - n_low) / 2)
                self.logger.info(
                    'NOT SATISFIED ' + info + '. low=' + str(n_low) + ', high=' + str(n_high) + ', next n=' + str(n)
                )
                if n <= n_low:
                    # Always take the lower dimension & higher grid density
                    if (n_low < n_high):
                        n_high = n_low
                        n = n_low
                    else:
                        self.logger.info(
                            'Already reached minimum n = ' + str(n_low)
                            + ', optimal not found for target grid density ' + str(target_grid_density)
                            + ', final grid density "' + str(measure) + '"=' + str(measure_value)
                            + ', details: ' + str(self.grid_density)
                        )
                        pca_fit_optimal = pca_fit_optimal
                        break
                continue
        return pca_fit_optimal

    def fit(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            # Model dependent interpretation, or ignore if not relevant for specific model
            # For example, can mean how many clusters, or how many PCA components, or how many to sample
            # in a discrete Fourier transform, etc.
            n_components = 2,
            # for fine-tuning already trained clusters, thus no need to start all over again
            # useful for clusters of more than 1,000,000 points for example, where starting
            # again means another half day of fit training
            start_centers: np.ndarray = None,
            # pass through mode
            test_mode = False,
    ) -> dict:
        if X_labels is None:
            X_labels = list(range(len(X)))
        try:
            self.__lock.acquire_mutexes(
                id = 'fit',
                mutexes = [self.__mutex_model],
            )
            return self.__fit(
                X = X,
                X_labels = X_labels,
                X_full_records = X_full_records,
                n_components = n_components,
            )
        finally:
            self.__lock.release_mutexes(mutexes=[self.__mutex_model])

    def __fit(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            # Model dependent interpretation, or ignore if not relevant for specific model
            # For example, can mean how many clusters, or how many PCA components, or how many to sample
            # in a discrete Fourier transform, etc.
            n_components = 2,
            # pass through mode
            test_mode = False,
    ):
        assert type(X) is np.ndarray, 'Wrong type X "' + str(type(X)) + '"'
        if X_labels is None:
            X_labels = list(range(len(X)))
        pca = PCA(n_components=n_components)

        # apply principal component analysis to the embeddings table
        # df_reduced = pca.fit_transform(embeddings_df[embeddings_df.columns[:-2]])
        x_reduced = pca.fit_transform(pd.DataFrame(X))

        self.model_params_ready = False

        # Keep model parameters
        self.model_train_total_iterations = None
        self.model_n_components_or_centers = n_components
        # don't keep large tensors in memory
        # self.X = np.array(X)
        self.X_transform = x_reduced
        self.X_labels = X_labels
        self.X_full_records = X_full_records
        self.model_principal_components = pca.components_
        self.model_centroid = np.mean(X, axis=0)
        # not applicable to PCA
        self.model_centers = np.array([])

        self.__check_consistency(principal_components=self.model_principal_components)

        # Estimate back from principal components
        self.X_inverse_transform = self.__inverse_transform(
            x_transform = self.X_transform,
        )
        self.X_grid_vectors, self.X_grid_numbers = self.__calc_pca_grid(
            x_pca = self.X_transform,
        )

        X_lengths = np.sum((X * X), axis=-1) ** 0.5
        X_inverse_lengths = np.sum((self.X_inverse_transform * self.X_inverse_transform), axis=-1) ** 0.5

        # Equivalent to cluster "inertia"
        self.distance_error = np.sum((X - self.X_inverse_transform) ** 2, axis=-1) ** 0.5
        self.distance_error = self.distance_error / X_lengths
        self.distance_error_mean = np.mean(self.distance_error)

        self.angle_error = np.sum(X * self.X_inverse_transform, axis=-1) / (X_lengths * X_inverse_lengths)
        self.angle_error_mean = np.mean(self.angle_error)

        self.grid_density = np.zeros(2**self.X_grid_vectors.shape[-1])
        for i in range(len(self.grid_density)):
            self.grid_density[i] = np.sum(1 * (self.X_grid_numbers == i))
        self.grid_density_mean = np.mean(self.grid_density)

        self.model_params_ready = True

        return self.model_to_b64json(numpy_to_base64_str=False)

    def fine_tune(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            n_components: int = None,
            # pass through mode
            test_mode = False,
    ) -> dict:
        # TODO How to fine tune?
        self.logger.warning('Fine tuning for PCA not yet implemented')
        res = self.fit(
            X = X,
            X_labels = X_labels,
            X_full_records = X_full_records,
            n_components = n_components,
        )
        return res

    # Recover estimate of original point from PCA compression
    def inverse_transform(
            self,
            X: np.ndarray,
    ) -> np.ndarray:
        try:
            self.__lock.acquire_mutexes(
                id = 'inverse_transform',
                mutexes = [self.__mutex_model],
            )
            return self.__inverse_transform(
                x_transform = X,
            )
        finally:
            self.__lock.release_mutexes(mutexes=[self.__mutex_model])

    def __inverse_transform(
            self,
            x_transform: np.ndarray,
    ) -> np.ndarray:
        self.__check_consistency(principal_components=self.model_principal_components)
        if x_transform.ndim == 1:
            x_transform = x_transform.reshape((1, x_transform.shape[0]))
        x_estimated = np.zeros(shape=(len(x_transform), self.model_centroid.shape[-1]))
        x_estimated += self.model_centroid
        # TODO how to not use loop?
        for i in range(len(x_estimated)):
            for k in range(len(self.model_principal_components)):
                x_estimated[i] += x_transform[i][k] * self.model_principal_components[k]
        return x_estimated

    # Get PCA values of arbitrary points
    def transform(
            self,
            X: np.ndarray,
    ) -> np.ndarray:
        try:
            self.__lock.acquire_mutexes(
                id = 'transform',
                mutexes = [self.__mutex_model],
            )
            return self.__transform(X=X)
        finally:
            self.__lock.release_mutexes(mutexes=[self.__mutex_model])

    def __transform(
            self,
            X: np.ndarray,
    ) -> np.ndarray:
        self.__check_consistency(principal_components=self.model_principal_components)
        assert X.shape[-1] == self.model_centroid.shape[0], (
                'Inconsistent shapes X ' + str(X.shape) + ', centroid ' + str(self.model_centroid.shape))
        X_remainder = X - self.model_centroid
        pca_coef = np.zeros(shape=(len(X), len(self.model_principal_components)))
        for k in range(len(self.model_principal_components)):
            if k > 0:
                pca_coef_1 = pca_coef[:, k - 1]
                pca_coef_1 = pca_coef_1.reshape((len(pca_coef_1), 1))
                # print('X_remainder', X_remainder)
                # print('pca coef k-1', pca_coef_1)
                # print('pca components k-1', pca_components[k-1])
                X_remainder = X_remainder - pca_coef_1 * self.model_principal_components[k - 1]
            pca_coef[:, k] = np.matmul(X_remainder, self.model_principal_components[k].transpose())
            # print(k, pca_coef)
        return pca_coef

    def __calc_pca_grid(
            self,
            x_pca: np.ndarray,
    ):
        grids = np.zeros(shape=(len(x_pca), self.model_n_components_or_centers))
        grid_numbers = np.zeros(len(x_pca))
        for i in range(self.model_n_components_or_centers):
            # Build +/- binary for each pca component. For example if we have 2 rows, with 3 PCA components
            # [
            #   [  1.2, -9.5, 0.8 ],
            #   [ -4.0, -8.5, -0.4 ],
            # ]
            # Then we would have grids
            # [
            #   [  1, -1, 1 ],
            #   [ -1, -1, -1 ],
            # ]
            # Thus n components will have 2^n grids, for the case above with 3 components will have max 8 grids.
            # Since PCA is centered, we are somewhat assured of uniform layout, and not everything concentrated
            # in a few grids.
            grids[:,i] = 1 * (x_pca[:,i] >= 0)
            # big endian
            grid_numbers += (2**(self.model_n_components_or_centers-i-1)) * grids[:,i]
        return grids, grid_numbers

    def predict(
            self,
            X: np.ndarray,
            # can be cluster numbers to zoom into
            X_search_local_space: np.ndarray = None,
            labels_search_local_space: list = None,
            top_k = 5,
            return_full_record = False,
            use_grid = False,
    ):
        try:
            self.__lock.acquire_mutexes(
                id = 'predict',
                mutexes = [self.__mutex_model],
            )

            #
            # There are 2 possible approaches, after obtaining the PCA segment numbers & relevant reference vectors:
            #    1. Transform input vector to PCA transform, then compare with the reference PCA transforms
            #    2. Do not transform, use original vectors to compare. For now, we use this approach to skip the step
            #       of transforming the input vector.
            #

            X_pca = self.__transform(
                X = X,
            )

            if use_grid:
                #
                # STEP 1: Determine the grid segments of the texts/embeddings
                #
                grid, grid_numbers = self.__calc_pca_grid(
                    x_pca = X_pca,
                )
                self.logger.info('Predict X lies in grid numbers: ' + str(grid_numbers))

                #
                # STEP 2: Get the reference text encoding, standardized labels, data records for those grid segments only
                #
                condition = False
                # recs_from_db = []
                for gn in np.unique(grid_numbers):
                    condition = condition | (self.X_grid_numbers == gn)
                # TODO This can be obtained by querying underlying DB, so we don't need to keep in RAM
                X_subset = self.X_transform[condition]
                labels_subset = np.array(self.X_labels)[condition].tolist()
                if return_full_record:
                    assert self.X_full_records is not None, 'Cannot return full records'
                    full_records_list = np.array(self.X_full_records)[condition].tolist()
                else:
                    full_records_list = None
                # data_records_subset = [r for i, r in enumerate(self.model_compression_data_records) if condition[i]]
            else:
                X_subset = self.X_transform
                labels_subset = self.X_labels
                full_records_list = self.X_full_records

            #
            # STEP 3: Call __predict()
            #
            pred_labels_or_records, pred_probs = self.predict_standard(
                X = X_pca,
                ref_X = X_subset,
                ref_labels = labels_subset,
                ref_full_records = full_records_list,
                top_k = top_k,
                return_full_record = return_full_record,
            )
            return pred_labels_or_records, pred_probs
        finally:
            self.__lock.release_mutexes(mutexes=[self.__mutex_model])

    def model_to_b64json(
            self,
            numpy_to_base64_str = False,
            dump_to_b64json_str = False,
    ):
        base_model_dict = super().model_to_b64json(
            numpy_to_base64_str = numpy_to_base64_str,
            dump_to_b64json_str = False,
        )
        if dump_to_b64json_str:
            return self.base64.encode(b=json.dumps(base_model_dict).encode(encoding='utf-8'))
        else:
            return base_model_dict

    def load_model_from_b64json(
            self,
            model_b64json,
    ):
        _ = super().load_model_from_b64json(model_b64json=model_b64json)
        return


if __name__ == '__main__':
    from fitxf.math.lang.encode.LangModelPt import LangModelPt as LmPt
    texts = [
        "Let's have coffee", "Free for a drink?", "How about Starbucks?",
        "I am busy", "Go away", "Don't disturb me",
        "Monetary policies", "Interest rates", "Deposit rates",
    ]
    lmo = LmPt(cache_folder=Env().MODELS_PRETRAINED_DIR)

    embeddings = lmo.encode(content_list=texts, return_tensors='np')

    # use the function create_pca_plot to
    fitter = FitXformPca(logger=Logging.get_default_logger(log_level=logging.INFO, propagate=False))
    pca_fit_optimal = fitter.fit_optimal(X=embeddings)
    x_fit = pca_fit_optimal['X_transform']

    fitter.create_scatter_plot2d(
        x_transform = x_fit,
        labels_list = texts,
        show = True,
    )

    x = np.array([[1,2,3], [3,2,1], [-1,-2,-2], [-3,-4,-2]])
    x_pca = fitter.fit(X=x, X_labels=['+', '+', '-', '-'], n_components=1)
    print(x_pca)
    print(fitter.predict(X=np.array([[9,9,8], [-55,-33,-55]]), use_grid=False))
    print(fitter.predict(X=np.array([[9,9,8], [-55,-33,-55]]), use_grid=True))

    exit(0)



import logging
import os
import re
import platform
import numpy as np
from fitxf.math.fit.utils.PatternSearch import PatternSearch
from fitxf.math.utils.Pandas import Pandas
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Profile import Profiling


class PatternSearchUnitTest:

    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(self):
        patsearch = PatternSearch(logger=self.logger)

        x = np.arange(20) % 10
        for seq, exp_matches in [
            (np.array([1, 2, 3, 4]), np.array([1, 11])),
            (np.array([9, 10, 11]), np.array([9])),
            # (np.array([1, 3, 5]), []),
        ]:
            match_idxs = patsearch.math_utils.match_template(x=x, seq=seq)
            assert np.sum((np.array(match_idxs) - exp_matches)**2) < 0.0000000001, \
                'Match indexes ' + str(match_idxs) + ' not ' + str(exp_matches)

        dth, cth = 0.9, 0.9
        test_data = [
            # No repeats
            (dth, cth, "don't destroy this genuine sentence", ""),
            (dth, cth, "another valid sentence with no LLM hysterics", ""),
            # Have repeats
            (dth, 0.6, "zz zz zz abedeee abedeee abedeeee", "abedeee abedeee abedeee"),
            (1.0, cth, "Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1",
             "Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 "),
            (1.0, cth, "Test 222222 Test 222222 Test 222222 Test 222222 Test 222222 Test 222222 Tes",
             "Test 222222 Test 222222 Test 222222 Test 222222 Test 222222 Test 222222 "),
            (dth, 0.6, "bread rice wine beer wine beer", "wine beer wine beer"),
            (dth, 0.6, "LLM hysterics ||| ||| ||| ||| ||| ||| ||", "||| ||| ||| ||| ||| ||| "),
        ]
        for den_thr, cov_thr, s, exp_repeats in test_data:
            res = patsearch.find_repeat_sequences(
                x = np.array([ord(c) for c in s]),
                min_seq_len = 3,
                density_repeat_thr = den_thr,
                coverage_thr = cov_thr,
                hint_separators = np.array([ord(v) for v in [" "]]),
                stagnant_at_optimal_count_thr = 10,
                string_convert_result = True,
            )
            if res is not None:
                s_front, s_repeats, s_back = res[0]['s_parts']
            else:
                s_repeats = ''
            assert s_repeats == exp_repeats, 'Expected "' + str(exp_repeats) + '", got "' + str(s_repeats) + '"'

        for i, (min_c_len, hint_sep, den_thr, cov_thr, s, exp_res) in enumerate([
            #
            # No repeats
            #
            (8, [".", ":", ";"], 0.5, 0.5, "LLMs go hysterical always", None),
            (8, [".", ":", ";"], 0.5, 0.5, "LLMs go hysterical every time, what nonsense from AI", None),
            #
            # Optimal repeats
            #
            (2, [" "], 0.6, 0.5, "AI LLMs 7778888 LLMs 777---- LLMs 777++++ LLMs 7778888 xx LLMs 777**** LLMs 777^^^^ ...",
             {
                 'seq_len': 8, 'prefix': "LLMs 777", 'indexes': [3, 16, 29, 42, 58, 71],
                 'seq_list': [2, 13, 12, 11, 10, 9, 8],
                 'break_reason': 'density/coverage condition met',
             }),
            #
            # Have repeats
            #
            (8, [".", ":", ";"], 0.6, 0.6, "LLMs go hysterical: LLMs go hysterical: LLMs go hysterical: LLMs g",
             {
                 'seq_len': 19, 'prefix': " LLMs go hysterical", 'indexes': [19, 39], 'seq_list': [8, 20],
                 'break_reason': 'density/coverage condition met',
             }),
            (16, [".", ":", ";"], 0.5, 0.5, "LLMs go hysterical: LLMs go hysterical: LLMs go hysterical: LLMs g",
             {
                 'seq_len': 16, 'prefix': " LLMs go hysteri", 'indexes': [19, 39], 'seq_list': [16],
                 'break_reason': 'density/coverage condition met',
             }),
            (8, [".", ":", ";"], 0.3, 0.5, "LLMs go hy111111: LLMs go hy222222222222: LLMs go hy33333333: LLMs g",
             {
                 'seq_len': 10, 'prefix': " LLMs go h", 'indexes': [17, 41], 'seq_list': [8, 24, 16, 12, 10, 11],
                 'break_reason': 'density/coverage condition met',
             }),
            # With space as hint separator
            (8, [".", ":", ";", " "], 0.9, 0.8, "LLMs go hysterical: LLMs go hysterical: LLMs go hysterical: L",
             {
                 'seq_len': 18, 'prefix': "LLMs go hysterical", 'indexes': [0, 20, 40], 'seq_list': [8, 20],
                 'break_reason': 'density/coverage condition met',
             }),
        ]):
            res = patsearch.find_repeat_sequences(
                x = np.array([ord(c) for c in s]),
                min_seq_len = min_c_len,
                hint_separators = np.array([ord(c) for c in hint_sep]),
                density_repeat_thr = den_thr,
                coverage_thr = cov_thr,
                stagnant_at_optimal_count_thr = 100,
                string_convert_result = True,
            )
            if exp_res is None:
                assert res is None
            else:
                assert res is not None, '#' + str(i) + ' result None for "' + str(s) + '"'
                pfx, sq_len, sq_lst, idxs, br = \
                    res[0]['prefix'], res[0]['seq_len'], res[0]['seq_list'], res[0]['indexes'], res[0]['break_reason']
                exp_pfx, exp_sq_len, exp_sq_lst, exp_idxs, exp_br = \
                    exp_res['prefix'], exp_res['seq_len'], exp_res['seq_list'], exp_res['indexes'], exp_res['break_reason']
                assert re.match(pattern=exp_pfx, string=pfx), \
                    '#' + str(i) + ' Expected prefix "' + str(exp_pfx) + '*" but got "' + str(pfx) + '"'
                assert sq_len >= exp_sq_len, \
                    '#' + str(i) + ' Expected >= seq len ' + str(exp_sq_len) + ' but got ' + str(sq_len)
                # assert sq_lst == exp_sq_lst, \
                #     '#' + str(i) + ' Expected seq list ' + str(exp_sq_lst) + ' but got ' + str(sq_lst)
                assert idxs == exp_idxs, \
                    '#' + str(i) + ' Expected indexes ' + str(exp_idxs) + ' but got ' + str(idxs)
                assert br == exp_br, \
                    '#' + str(i) + ' Expected break reason "' + str(exp_br) + '" but got "' + str(br) + '"'

        long_string = ""
        for i in range(10):
            long_string += " ".join([str(v) for v in np.arange(20).tolist()]) + " <RANDOM>" \
                           + str(np.random.random(5)) + '</RANDOM>.\n'
        long_string = long_string + " ".join([str(v) for v in np.arange(10).tolist()])
        res = patsearch.find_repeat_sequences(
            x = np.array([ord(c) for c in long_string]),
            min_seq_len = 11,
            hint_separators = np.array([ord(c) for c in ["\n"]]),
            stagnant_at_optimal_count_thr = 100,
            string_convert_result = True,
            density_repeat_thr = 0.5,
            coverage_thr = 0.5,
        )
        pfx, idxs, sqlst = res[0]["prefix"], res[0]["indexes"], res[0]["seq_list"]
        assert res[0]["prefix"] in [
            "0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 <RANDOM>[0.",
            "0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 <RANDOM>[0",
        ], 'Got ' + str(pfx)
        assert res[0]["indexes"] == [0, 125, 250, 375, 500, 625, 750, 875, 1000, 1125], 'Got instead ' + str(idxs)
        # assert res[0]["seq_list"] == [11, 125, 68, 40, 67, 50, 66, 57, 65, 60, 64, 62, 63, 61], 'Got ' + str(sqlst)

        # Quiet during speed test
        self.logger.setLevel(level=logging.ERROR)
        os.environ["ENABLE_SLOW_LOGGING"] = "false"
        obj = PatternSearch(
            logger = self.logger,
        )
        time_records = []
        prf = Profiling(logger=self.logger)
        for i in range(200):
            start = prf.start()
            res = obj.find_repeat_sequences(
                x = np.array([
                    ord(c) for c in "Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1 Test 1"]
                ),
                min_seq_len = 4,
                hint_separators = np.array([ord(c) for c in ["1", " "]]),
                string_convert_result = True,
            )
            ms = round(1000 * prf.get_time_dif_secs(start=start, stop=prf.stop()), 3)
            time_records.append(ms)
            print('#' + str(i) + ': ' + str(res)) if i%100==0 else 1

        # Noisy back
        self.logger.setLevel(level=logging.INFO)
        pf = platform.platform()
        pf_info = pf.split("-")
        processor = pf_info[2]
        if processor in ['x86_64']:
            avg_ms_thr = 32
        else:
            avg_ms_thr = 6
        self.logger.info(
            'Platform processor "' + str(processor) + '", "' + str(pf) + '": ' + str(pf_info)
            + ', avg ms thr ' + str(avg_ms_thr)
        )

        avg_ms = np.mean(np.array(time_records))
        self.logger.info('Average time ' + str(avg_ms) + ', time records: ' + str(time_records))
        assert avg_ms < avg_ms_thr, 'FAIL Average ms ' + str(avg_ms) + ' < ' + str(avg_ms_thr) + 'ms'

        self.logger.info('ALL TESTS PASSED')
        return


if __name__ == '__main__':
    Pandas.increase_display()
    lgr = Logging.get_default_logger(log_level=logging.DEBUG, propagate=False)
    PatternSearchUnitTest(logger=lgr).test()
    exit(0)



import logging
import datetime
import threading


class Profiling:

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        self.time_records = {}
        self.__mutex = threading.Lock()
        return

    def start(self):
        return datetime.datetime.now()

    def stop(self):
        return datetime.datetime.now()

    def get_time_dif_secs(
            self,
            start,
            stop,
            decimals=10,
    ):
        diftime = (stop - start)
        diftime = round(diftime.days*86400 + diftime.seconds + diftime.microseconds / 1000000, decimals)
        return diftime

    def start_time_profiling(
            self,
            id,
            force_reset = False,
            max_records = 1000,
    ):
        try:
            self.__mutex.acquire()

            dt_now = self.start()
            new_record = {'start': dt_now, 'stop': dt_now, 'dif_secs': -1.}

            if force_reset or (id not in self.time_records):
                self.time_records[id] = [new_record]
            else:
                is_non_empty = len(self.time_records[id]) >= 1
                # Means stuck from previous measurement, never call record after start
                if is_non_empty and (self.time_records[id][-1]['dif_secs']) < 0.:
                    # self.logger.warning('Did not call record_time_profiling() after start')
                    self.time_records[id][-1] = new_record
                    self.logger.warning('Previous time record "' + str(id) + '" no end time, replace with new record')
                else:
                    self.time_records[id].append(new_record)
                    self.logger.debug('Append new time recording to "' + str(id) + '"')

            if len(self.time_records[id]) > max_records:
                self.time_records[id] = self.time_records[id][1:]
        finally:
            self.__mutex.release()

    def record_time_profiling(
            self,
            id,
            msg,
            decimals = 10,
            logmsg = False,
    ):
        try:
            self.__mutex.acquire()
            self.time_records[id][-1]['stop'] = self.stop()
            dur_s = self.get_time_dif_secs(
                start = self.time_records[id][-1]['start'],
                stop  = self.time_records[id][-1]['stop'],
                decimals = decimals,
            )
            self.time_records[id][-1]['dif_secs'] = dur_s

            # Clean up those with no stop times
            # self.time_records[id] = [rec for rec in self.time_records[id] if rec['dif_secs'] > 0.]

            dur_cum_s = sum([rec['dif_secs'] for rec in self.time_records[id]])
            mean_secs = round(dur_cum_s / len(self.time_records[id]), decimals)
            rps_mean = round(1/(mean_secs+0.0000000001), 3)
            if logmsg:
                self.logger.info(
                    '[' + str(id) + '] Profile speed: ' + str(msg) + ' took ' + str(dur_s)
                    + 's, cumulative mean ' + str(mean_secs) + 's (' + str(rps_mean) + ' rps) from total records '
                    + str(len(self.time_records[id]))
                )
            return dur_s, dur_cum_s
        finally:
            self.__mutex.release()


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    import time
    pf = Profiling()
    def f():
        time.sleep(0.01)

    for i in range(10):
        pf.start_time_profiling(id='test', max_records=5)
        pf.start_time_profiling(id='test', max_records=5)
        pf.start_time_profiling(id='test', max_records=5)

        f()
        pf.record_time_profiling(
            id = 'test',
            msg = str(i),
            logmsg = True,
        )

    exit(0)



import logging
import os
import time
from datetime import datetime


class FileUtils:

    def __init__(
            self,
            filepath,
            logger = None,
    ):
        self.filepath = filepath
        self.logger = logger if logger is not None else logging.getLogger()

        self.last_modified_time = self.get_last_modified_time()
        self.logger.info(
            'File "' + str(self.filepath) + '", last modified time "' + str(self.last_modified_time)
        )
        return

    def update_last_modified_time(self):
        self.last_modified_time = self.get_last_modified_time()

    def get_last_modified_time(self):
        t = os.path.getmtime(self.filepath)
        last_modified_time = datetime.fromtimestamp(t)
        self.logger.debug(str(self.filepath) + ': file last modified time "' + str(last_modified_time))
        return last_modified_time

    def is_file_modified(self):
        t_delta = self.get_last_modified_time() - self.last_modified_time
        diff_secs = t_delta.days*86400 + t_delta.seconds + t_delta.microseconds/1000000
        self.logger.debug(str(self.filepath) + ': difference between last modified times = ' + str(diff_secs) + 's')
        return diff_secs > 0

    def read_text_file(
            self,
            encoding = 'utf-8',
            throw_exception = True,
    ):
        try:
            fh = open(self.filepath, 'r', encoding=encoding)
        except IOError as e:
            errmsg = 'Cannot open file [' + str(self.filepath) + ']. ' + str(e)
            self.logger.error(errmsg)
            if throw_exception:
                raise Exception(errmsg)
            else:
                return []

        lines = [row.rstrip() for row in fh]

        fh.close()
        return lines


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    filepath = 'filetest.txt'
    fu = FileUtils(filepath=filepath)

    while True:
        lns = fu.read_text_file()
        # [print(ln) for ln in lns]
        print(fu.get_last_modified_time())
        modified = fu.is_file_modified()
        if modified:
            break
        time.sleep(5)



import logging
import numpy as np
from fitxf.math.fit.cluster.Cluster import Cluster
from fitxf.math.fit.utils.TensorUtils import TensorUtils
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Profile import Profiling


#
# Idea is to calculate a matrix against itself for dot product
# Then we retrieve row by row, max cluster by row, until all points are covered
#
class ClusterCosine(Cluster):

    def __init__(
            self,
            logger = None,
    ):
        super().__init__(
            logger = logger,
        )
        self.tensor_utils = TensorUtils(logger=self.logger)
        self.profiler = Profiling(logger=self.logger)
        return

    def kmeans(
            self,
            x: np.ndarray,
            n_centers: int,
            x_labels: list = None,
            # for fine-tuning already trained clusters, thus no need to start all over again
            # useful for clusters of more than 1,000,000 points for example, where starting
            # again means another half day of fit training
            start_centers: np.ndarray = None,
            km_iters = 100,
            converge_diff_thr = 0.00001,
            # pass through mode
            test_mode = False,
    ):
        if test_mode:
            raise Exception('Test mode not supported')

        assert km_iters > 0
        start_time = self.profiler.start()

        if x_labels is None:
            x_labels = list(range(len(x)))

        self.logger.info('Start cosine cos clustering x of shape ' + str(x.shape))
        # For cosine similarity, it makes no sense not to normalize first
        x_normalized = self.tensor_utils.normalize(x=x, return_tensors='np')

        centroid_move_changes = []
        if start_centers is None:
            # randomly pick n unique points
            last_centers = np.unique(x_normalized, axis=0)[0:n_centers]
        else:
            self.logger.info('Using user provided start centers of shape ' + str(start_centers.shape))
            assert start_centers.shape[-1] == x.shape[-1], \
                'Last dim lengths not equal. Start centers shape ' + str(start_centers.shape) + ', x ' + str(x.shape)
            assert start_centers.ndim == x.ndim, \
                'Dimensions not equal, start centers shape ' + str(start_centers.shape) + ', x ' + str(x.shape)
            assert len(start_centers) == n_centers, \
                'N centers ' + str(n_centers) + ' not same length with start centers ' + str(start_centers.shape)
            last_centers = start_centers

        last_cluster_numbers = None
        last_clusters = None
        total_iters = 0
        for iter in range(km_iters):
            self.logger.info('Starting iteration #' + str(iter+1) + '...')
            # Closest centers for all points
            result_ordered, mdot_ordered = self.tensor_utils.dot_sim(
                x = x_normalized,
                ref = last_centers,
            )
            x_new_cluster_numbers = result_ordered[:,0].tolist()
            x_new_clusters = []
            for i in range(n_centers):
                i_cluster = [idx for (idx, clstr_no) in enumerate(x_new_cluster_numbers) if clstr_no == i]
                x_new_clusters.append(i_cluster)
            self.logger.debug('Result/mdot ordered ' + str(result_ordered) + ', ' + str(mdot_ordered))
            self.logger.debug(
                'N centers required ' + str(n_centers) + ', cluster numbers: ' + str(x_new_cluster_numbers)
            )

            # update new centroids
            updated_cluster_numbers, updated_centers = self.get_cluster_numbers_and_centroids(
                x = x_normalized,
                clusters = x_new_clusters,
            )
            assert updated_cluster_numbers.tolist() == x_new_cluster_numbers, \
                'Consistency off from updated cluster numbers ' + str(list(zip(updated_cluster_numbers, x_new_cluster_numbers)))
            assert updated_centers.shape == last_centers.shape, \
                'Centers shape not same after update ' + str(updated_centers.shape) + ' != ' + str(last_centers.shape)
            # it is easier to do Euclidean distance changes of last centers to updated centers
            dist_movements = np.sum((updated_centers - last_centers) ** 2, axis=-1) ** 0.5
            avg_dist_movements = np.mean(dist_movements)
            centroid_move_changes.append(avg_dist_movements)


            last_cluster_numbers = updated_cluster_numbers
            last_centers = updated_centers
            last_clusters = x_new_clusters

            if len(centroid_move_changes) >= 2:
                move_diff = np.abs(centroid_move_changes[-2] - centroid_move_changes[-1])
                move_ratio = 100 * np.abs(centroid_move_changes[-1] / centroid_move_changes[0])
                converge_cond = move_diff < converge_diff_thr
                self.logger.info(
                    'Movement diff ' + str(move_diff) + ', move ratio with initial '  + str(move_ratio)
                    + '%. Converge condition = ' + str(converge_cond)
                )
            else:
                converge_cond = False

            self.logger.info('Done iteration #' + str(iter+1) + ', converged = ' + str(converge_cond))
            total_iters = iter
            if converge_cond:
                break

        diff_secs = self.profiler.get_time_dif_secs(start=start_time, stop=self.profiler.stop(), decimals=4)
        self.logger.info(
            'Total time taken for kmeans clustering data shape ' + str(len(x_normalized.shape))
            + ' to ' + str(n_centers) + ' = ' + str(diff_secs) + 's.'
        )

        additional_info = self.derive_additional_cluster_info(
            x = x_normalized,
            n_centers = n_centers,
            cluster_centers = last_centers,
            cluster_labels = last_cluster_numbers,
            metric = 'cosine',
        )
        if x_labels is not None:
            cluster_label_to_labelsori = self.map_centers_to_original_labels(
                labels_original = x_labels,
                labels_cluster = last_cluster_numbers.tolist(),
            )
        else:
            cluster_label_to_labelsori = None
        return {
            'total_iterations': total_iters,
            'n_centers': n_centers,
            'clusters': last_clusters,
            'cluster_centers': last_centers,
            # correspond to the index of the "centroids"
            'cluster_labels': last_cluster_numbers,
            'cluster_label_to_original_labels': cluster_label_to_labelsori,
            'centers_median': additional_info['centers_median'],
            'inner_radiuses': additional_info['inner_radiuses'],
            'cluster_sizes': additional_info['cluster_sizes'],
            # estimate as inner_radiuses
            'points_inertia': np.mean(additional_info['inner_radiuses']),
        }

    def get_cluster_numbers_and_centroids(
            self,
            # e.g.
            # [[1,2,0], [0,4,5], [6,3,4], [5,5,5], [0,9,8]]
            x,
            # list of clusters by x indexes e.g. [[0,1], [2,3], [4]]
            clusters: list,
    ):
        len_x = len(x)
        center_shape = list(x.shape[1:])
        self.logger.info(
            'Clusters len ' + str(len(clusters)) + ', x shape ' + str(x.shape)
            + ', center without first dim shape ' + str(center_shape)
        )
        center_shape_1d = 1
        for i in center_shape:
            center_shape_1d *= i

        new_centers = []
        empty_centers_indexes = []
        cluster_item_counts = []
        cluster_numbers = np.array([-1]*len_x)
        for i, clstr in enumerate(clusters):
            # assert len(clstr) > 0, 'Empty cluster at ' + '#' + str(i) + ', cluster ' + str(clstr)
            # It can happen that no points are in cluster i, when for example user provided the start centers
            # during iteration, thus we must simply pick another suitable random point
            if len(clstr) == 0:
                self.logger.warning('No points attached to cluster center i=' + str(i))
                empty_centers_indexes.append(i)
                new_centers.append(None)
                cluster_item_counts.append(0)
            else:
                select = np.array([False]*len_x)
                for item in clstr:
                    select[item] = True
                center = x[select].mean(axis=0)
                new_centers.append(center.tolist())
                cluster_numbers[np.array(clstr)] = i
                cluster_item_counts.append(len(clstr))

        if len(empty_centers_indexes) > 0:
            self.logger.info(
                'Empty clusters exist total ' + str(len(empty_centers_indexes))
                + ' at center indexes ' + str(empty_centers_indexes)
            )
            np_cluster_item_counts = np.array(cluster_item_counts)
            # Now we know which index has most points (last index), and which has least (index 0)
            np_cluster_item_counts_sorted = np.argsort(np_cluster_item_counts)
            # We split the clusters with most points if there are empty clusters
            clusterno_to_split = np_cluster_item_counts_sorted[-1]
            cluster_to_split = clusters[clusterno_to_split]
            self.logger.info(
                'Empty clusters during train, using cluster at index #' + str(clusterno_to_split) + ' to split with '
                + str(len(cluster_to_split)) + ' elements ' + str(cluster_to_split)
            )

            for i in empty_centers_indexes:
                j = np.random.randint(low=0, high=len(cluster_to_split), size=1)[0]
                point_index = cluster_to_split[j]
                random_new_center = np.reshape(x[point_index], newshape=center_shape)
                # Make sure add a very small random number
                random_new_center += np.reshape(np.random.rand(center_shape_1d), newshape=center_shape) / 1000000
                new_centers[i] = random_new_center.tolist()
                self.logger.info(
                    'Reassigning empty cluster center ' + str(i) + ' a random point taking mean from index '
                    + str(point_index) + '.'
                )
        self.logger.info('Final centers length ' + str(len(new_centers)))
        self.logger.debug('Final clusters ' + str(cluster_numbers))
        return np.array(cluster_numbers), np.array(new_centers)

    def cluster_angle(
            self,
            # vectors
            x: np.ndarray,
            n_clusters: int,
            max_iter: int = 10,
            start_min_dist_abs: float = 0.8,
    ):
        start_time = self.profiler.start()

        x_norm = self.tensor_utils.normalize(
            x = x,
            return_tensors = 'np',
        )
        ref_norm = x_norm.transpose()
        l = len(x)
        ref_array = np.array(list(range(l)) * l).reshape(l,l) + 1

        # can be negative which means in the opposite direction from reference
        m_dot = np.matmul(x_norm, ref_norm)

        min_dist_abs = start_min_dist_abs
        clusters = []
        # we presume last move is positive, thus to tighten or increase clusters
        last_move = 0.2
        for iter in range(max_iter):
            clusters = []
            # Will give an array of 0's (fail condition) & 1's (meet condition)
            # [[1 1 0 0]
            #  [1 1 0 0]
            #  [0 0 1 1]
            #  [0 0 1 1]]
            meet_condition = 1 * (np.abs(m_dot) > np.abs(min_dist_abs))
            # Give and array of the clusters by row, 0 means None, item indexing starts from 1
            # [[1 2 0 0]
            #  [1 2 0 0]
            #  [0 0 3 4]
            #  [0 0 3 4]]
            iter_clusters = ref_array * meet_condition
            self.logger.debug(
                'Iter #' + str(iter) + ': meet condition\n' + str(meet_condition) + '\nclusters\n' + str(iter_clusters)
            )
            for i_cluster in iter_clusters:
                row_cluster = set([v for v in i_cluster if v > 0])
                self.logger.debug('Row cluster: ' + str(row_cluster))
                for existing_cluster in clusters:
                    # Keep removing items that already found clusters
                    row_cluster = set(row_cluster).difference(set(existing_cluster))
                # remaining items if non-empty & not already in clusters, then is a new cluster
                if row_cluster and (row_cluster not in clusters):
                    clusters.append(row_cluster)
            self.logger.debug('Clusters at iteration #' + str(iter) + ': ' + str(clusters))
            cur_len = len(clusters)
            if cur_len > n_clusters:
                # if last move different direction, move opposite direction a bit less, else slow down a little same dir
                move = -last_move / 2 if last_move > 0 else last_move*0.9
                # need to decrease threshold distance to reduce clusters
                min_dist_abs_new = max(0.0001, min_dist_abs + move)
                last_move = move
            elif cur_len < n_clusters:
                # if last move different direction, move opposite direction a bit less, else slow down a little same dir
                move = -last_move / 2 if last_move < 0 else last_move*0.9
                # need to increase
                min_dist_abs_new = min(0.9999, min_dist_abs + move)
                last_move = move
            else:
                break
            self.logger.info(
                'Iter #' + str(iter) + ': Adjusted dist thr from ' + str(min_dist_abs)
                + ' to new value ' + str(min_dist_abs_new) + '. Cluster n=' + str(cur_len)
                + ', target n=' + str(n_clusters) + '.'
            )
            min_dist_abs = min_dist_abs_new
        # minus 1 back
        clusters_corrected = []
        for s in clusters:
            # minus 1 and also convert back to int from numpy type int32 or int64
            clusters_corrected.append([int(element-1) for element in s])

        union_all = set()
        for s in clusters_corrected:
            union_all = union_all.union(s)
        union_all = list(union_all)
        union_all.sort()
        assert union_all == list(range(l)), 'Union all ' + str(union_all)

        cluster_numbers, centroids = self.get_cluster_numbers_and_centroids(
            x = x_norm,
            clusters = clusters_corrected,
        )
        self.logger.info(
            'Done cluster: ' + str(clusters_corrected) + ', cluster numbers: ' + str(cluster_numbers)
            + ', centroids: ' + str(centroids)
        )

        diff_secs = self.profiler.get_time_dif_secs(start=start_time, stop=self.profiler.stop(), decimals=4)
        self.logger.info(
            'Total time taken for clustering cosine shape ' + str(len(x_norm.shape)) + ' to ' + str(n_clusters)
            + ' = ' + str(diff_secs) + 's.'
        )
        return {
            'clusters': clusters_corrected,
            # correspond to the index of the "centroids"
            'cluster_numbers': cluster_numbers,
            # index of the centroids is the cluster numbers
            'centroids': centroids,
            'dist_threshold': min_dist_abs,
        }


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    dlen = 10000
    ddim = 384
    n = int(dlen / 100)
    x = np.random.random(size=(dlen, ddim))
    # print(x)
    ca = ClusterCosine(logger=lgr)
    # res = ca.cluster_angle(x=x, n_clusters=n, max_iter=100, start_min_dist_abs = 0.9)
    # print('------------------------------------------')
    # print(res)
    res = ca.kmeans(x=x, n_centers=n, km_iters=100)
    print('++++++++++++++++++++++++++++++++++++++++++')
    # print(res)
    print('Cluster densities: ' + str([len(c) for c in res['clusters']]))
    exit(0)



import logging
import logging.handlers
import os
import sys
from datetime import datetime


class Logging:

    LOGGER_NAME_DEFAULT = 'fitxf'
    LOGGER_FILENAME_DEFAULT = 'fitxf'

    LOG_FORMAT_DEFAULT = '%(asctime)s: %(name)s: %(levelname)s: <%(filename)s> "%(funcName)s" line #%(lineno)d:\t%(message)s'

    @staticmethod
    def set_debug_config():
        logging.basicConfig(
            level = logging.DEBUG,
            format = Logging.LOG_FORMAT_DEFAULT,
        )

    @staticmethod
    def __set_log_level(
            logger,
            # can be string or number
            log_level,
    ):
        print('Setting log level to "' + str(log_level) + '"')
        if log_level == 'DEBUG':
            logging.basicConfig(level=logging.DEBUG)
            logger.setLevel(level=logging.DEBUG)
        elif log_level == 'INFO':
            logging.basicConfig(level=logging.INFO)
            logger.setLevel(level=logging.INFO)
        elif log_level == 'WARN':
            logging.basicConfig(level=logging.WARN)
            logger.setLevel(level=logging.WARN)
        elif log_level == 'ERROR':
            logging.basicConfig(level=logging.ERROR)
            logger.setLevel(level=logging.ERROR)
        else:
            logging.basicConfig(level=log_level)
            logger.setLevel(level=log_level)
        print('Log level to "' + str(logger.getEffectiveLevel()) + '"')
        return

    @staticmethod
    def get_logger_from_env_var():
        log_level = logging.DEBUG if os.environ["LOGGER_LEVEL"].lower() == "debug" else None
        log_level = logging.WARN if os.environ["LOGGER_LEVEL"].lower() == "warn" else log_level
        log_level = logging.ERROR if os.environ["LOGGER_LEVEL"].lower() == "error" else log_level
        log_level = logging.INFO if log_level is None else log_level

        if os.environ["LOGGER_TYPE"] == 'rotatingfile':
            return Logging.get_logger_rotating_file(
                logger_name = os.environ["LOGGER_NAME"],
                log_level = log_level,
                filename = os.environ["LOGGER_FILE_PATH"],
                max_bytes = int(os.environ["LOGGER_MAX_MB"]) * 1048576,
                backup_count = int(os.environ["LOGGER_BACKUP_COUNT"]),
                propagate = False,
            )
        else:
            return Logging.get_default_logger(
                log_level = log_level,
                propagate = False,
            )

    """
    Rotating file logger, so we don't need to manage deletion of old log files.
    It will rotate between the fixed number of files, and overwrite the old data.
    """
    @staticmethod
    def get_logger_rotating_file(
            logger_name,
            log_level,
            filename,
            log_format = LOG_FORMAT_DEFAULT,
            # default 100MB
            max_bytes = (1048576 * 100),
            # means max 7 files to rotate each of size <maxBytes>
            backup_count = 7,
            # do not propagate to other loggers (prevent duplicate logging) if False
            propagate = False,
    ):
        # assert os.path.exists(filename), 'File path "' + str(filename) + '" must exist'
        logger = logging.getLogger(logger_name)
        for h in logger.handlers:
            print('Remove handler "' + str(h.name) + '", type "' + str(type(h)))
            logger.removeHandler(hdlr=h)

        Logging.__set_log_level(
            logger = logger,
            log_level = log_level,
        )
        # why logging library hasHandlers()==True, but inside handlers have nothing?
        print(
            'Handlers: ' + str(logger.handlers) + ' have handlers = ' + str(logger.hasHandlers())
            + ', log level "' + str(logger.getEffectiveLevel()) + '"'
        )

        handler = logging.handlers.RotatingFileHandler(
            filename = filename,
            maxBytes = max_bytes,
            backupCount = backup_count,
        )
        print(
            'Logger "' + str(logger_name) + ' added new RotatingFileHandler with filename "' + str(filename)
            + '", max bytes ' + str(max_bytes) + ', backup ' + str(backup_count)
        )
        formatter = logging.Formatter(log_format)
        handler.setFormatter(fmt=formatter)
        logger.addHandler(hdlr=handler)
        logger.propagate = propagate
        return logger

    @staticmethod
    def get_default_logger(
            log_level = logging.INFO,
            # do not propagate to other loggers (prevent duplicate logging) if False
            propagate = False,
            stream = 'stdout',
    ):
        # assert os.path.exists(filename), 'File path "' + str(filename) + '" must exist'
        logger = logging.getLogger(Logging.LOGGER_NAME_DEFAULT)
        for h in logger.handlers:
            print('Remove handler "' + str(h.name) + '", type "' + str(type(h)))
            logger.removeHandler(hdlr=h)

        Logging.__set_log_level(
            logger = logger,
            log_level = log_level,
        )

        formatter = logging.Formatter(Logging.LOG_FORMAT_DEFAULT)
        consoleHandler = logging.StreamHandler(
            stream = sys.stdout if stream in ('stdout',) else sys.stderr,
        )
        consoleHandler.setFormatter(formatter)
        logger.addHandler(consoleHandler)
        logger.propagate = propagate
        return logger


if __name__ == '__main__':
    def make_noise(lg, msg):
        lg.debug(msg)
        lg.info(msg)
        lg.warning(msg)
        lg.error(msg)
        lg.critical(msg)
        lg.fatal(msg)

    name = 'test_logger'
    lgr = Logging.get_logger_rotating_file(
        logger_name = name,
        log_level  = logging.INFO,
        log_format = Logging.LOG_FORMAT_DEFAULT,
        filename = 'test.log.' + str(datetime.now().strftime('%Y-%m-%d')) + '',
        max_bytes = (400),
        backup_count = 2,
    )
    for msg in ['moscow', 'astana']:
        make_noise(lg=lgr, msg=msg)

    lgr = Logging.get_logger_rotating_file(
        logger_name = name,
        log_level  = logging.INFO,
        log_format = Logging.LOG_FORMAT_DEFAULT,
        filename = 'test.log.' + str(datetime.now().strftime('%Y-%m-%d')) + '',
        max_bytes = (1000),
        backup_count = 2,
    )
    make_noise(lg=lgr, msg='melbourne')

    logging.critical('test logging without using logger')

    exit(0)
